# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Custom form widgets for WuttaFarm
"""

import json

import colander

from wuttaweb.db import Session
from wuttaweb.forms.schema import ObjectRef, WuttaSet, WuttaList
from wuttaweb.forms.widgets import NotesWidget


class AnimalTypeRef(ObjectRef):
    """
    Custom schema type for a
    :class:`~wuttafarm.db.model.animals.AnimalType` reference field.

    This is a subclass of
    :class:`~wuttaweb:wuttaweb.forms.schema.ObjectRef`.
    """

    @property
    def model_class(self):  # pylint: disable=empty-docstring
        """ """
        model = self.app.model
        return model.AnimalType

    def sort_query(self, query):  # pylint: disable=empty-docstring
        """ """
        return query.order_by(self.model_class.name)

    def get_object_url(self, obj):  # pylint: disable=empty-docstring
        """ """
        animal_type = obj
        return self.request.route_url("animal_types.view", uuid=animal_type.uuid)

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import AnimalTypeRefWidget

        kwargs["factory"] = AnimalTypeRefWidget
        return super().widget_maker(**kwargs)


class LogQuick(WuttaSet):

    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null

        return json.dumps(appstruct)

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import LogQuickWidget

        return LogQuickWidget(**kwargs)


class LogRef(ObjectRef):
    """
    Custom schema type for a
    :class:`~wuttafarm.db.model.log.Log` reference field.

    This is a subclass of
    :class:`~wuttaweb:wuttaweb.forms.schema.ObjectRef`.
    """

    @property
    def model_class(self):  # pylint: disable=empty-docstring
        """ """
        model = self.app.model
        return model.Log

    def sort_query(self, query):  # pylint: disable=empty-docstring
        """ """
        return query.order_by(self.model_class.message)

    def get_object_url(self, obj):  # pylint: disable=empty-docstring
        """ """
        log = obj
        return self.request.route_url(f"logs_{log.log_type}.view", uuid=log.uuid)


class FarmOSUnitRef(colander.SchemaType):

    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null

        return json.dumps(appstruct)

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import FarmOSUnitRefWidget

        return FarmOSUnitRefWidget(**kwargs)


class FarmOSRef(colander.SchemaType):

    def __init__(self, request, route_prefix, *args, **kwargs):
        self.values = kwargs.pop("values", None)
        super().__init__(*args, **kwargs)
        self.request = request
        self.route_prefix = route_prefix

    def get_values(self):
        if callable(self.values):
            self.values = self.values()
        return self.values

    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null

        # nb. keep a ref to this for later use
        node.model_instance = appstruct

        # serialize to PK as string
        return appstruct["uuid"]

    def deserialize(self, node, cstruct):
        if not cstruct:
            return colander.null

        # nb. deserialize to PK string, not dict
        return cstruct

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import FarmOSRefWidget

        if not kwargs.get("readonly"):
            if "values" not in kwargs:
                if values := self.get_values():
                    kwargs["values"] = values

        return FarmOSRefWidget(self.request, self.route_prefix, **kwargs)


class FarmOSRefs(WuttaSet):

    def __init__(self, request, route_prefix, *args, **kwargs):
        super().__init__(request, *args, **kwargs)
        self.route_prefix = route_prefix

    def serialize(self, node, appstruct):
        if not appstruct:
            return colander.null
        return appstruct

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import FarmOSRefsWidget

        return FarmOSRefsWidget(self.request, self.route_prefix, **kwargs)


class FarmOSAssetRefs(WuttaSet):

    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null

        return json.dumps(appstruct)

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import FarmOSAssetRefsWidget

        return FarmOSAssetRefsWidget(self.request, **kwargs)


class FarmOSLocationRefs(WuttaSet):

    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null

        return json.dumps(appstruct)

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import FarmOSLocationRefsWidget

        return FarmOSLocationRefsWidget(self.request, **kwargs)


class FarmOSQuantityRefs(WuttaSet):

    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null

        return json.dumps(appstruct)

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import FarmOSQuantityRefsWidget

        return FarmOSQuantityRefsWidget(**kwargs)


class FarmOSTaxonomyTerms(colander.SchemaType):
    """
    Schema type which can represent multiple taxonomy terms.
    """

    route_prefix = None

    def __init__(self, request, route_prefix=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        if route_prefix:
            self.route_prefix = route_prefix

    def serialize(self, node, appstruct):
        if not appstruct:
            return colander.null
        return appstruct

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import FarmOSTaxonomyTermsWidget

        return FarmOSTaxonomyTermsWidget(self.request, self.route_prefix, **kwargs)


class FarmOSEquipmentTypeRefs(FarmOSTaxonomyTerms):

    route_prefix = "farmos_equipment_types"


class FarmOSPlantTypes(colander.SchemaType):

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null

        return json.dumps(appstruct)

    def widget_maker(self, **kwargs):  # pylint: disable=empty-docstring
        """ """
        from wuttafarm.web.forms.widgets import FarmOSPlantTypesWidget

        return FarmOSPlantTypesWidget(self.request, **kwargs)


class LandTypeRef(ObjectRef):
    """
    Custom schema type for a
    :class:`~wuttafarm.db.model.land.LandType` reference field.

    This is a subclass of
    :class:`~wuttaweb:wuttaweb.forms.schema.ObjectRef`.
    """

    @property
    def model_class(self):  # pylint: disable=empty-docstring
        """ """
        model = self.app.model
        return model.LandType

    def sort_query(self, query):  # pylint: disable=empty-docstring
        """ """
        return query.order_by(self.model_class.name)

    def get_object_url(self, obj):  # pylint: disable=empty-docstring
        """ """
        land_type = obj
        return self.request.route_url("land_types.view", uuid=land_type.uuid)


class TaxonomyTermRefs(WuttaList):
    """
    Generic schema type for a field which can reference multiple
    taxonomy terms.
    """

    def serialize(self, node, appstruct):
        if not appstruct:
            return colander.null

        terms = []
        for term in appstruct:
            terms.append(
                {
                    "uuid": str(term.uuid),
                    "name": term.name,
                }
            )
        return terms


class EquipmentTypeRefs(TaxonomyTermRefs):

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import EquipmentTypeRefsWidget

        return EquipmentTypeRefsWidget(self.request, **kwargs)


class PlantTypeRefs(WuttaSet):
    """
    Schema type for Plant Types field (on a Plant Asset).
    """

    def serialize(self, node, appstruct):
        if not appstruct:
            return colander.null

        return [uuid.hex for uuid in appstruct]

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import PlantTypeRefsWidget

        model = self.app.model
        session = Session()

        if "values" not in kwargs:
            plant_types = (
                session.query(model.PlantType).order_by(model.PlantType.name).all()
            )
            values = [(pt.uuid.hex, str(pt)) for pt in plant_types]
            kwargs["values"] = values

        return PlantTypeRefsWidget(self.request, **kwargs)


class MaterialTypeRefs(colander.List):
    """
    Schema type for Material Types field (on a Material Asset).
    """

    def __init__(self, request):
        super().__init__()
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, node, appstruct):
        if not appstruct:
            return colander.null

        mtypes = []
        for mtype in appstruct:
            mtypes.append(
                {
                    "uuid": mtype.uuid.hex,
                    "name": mtype.name,
                }
            )
        return mtypes

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import MaterialTypeRefsWidget

        return MaterialTypeRefsWidget(self.request, **kwargs)


class SeasonRefs(WuttaSet):
    """
    Schema type for Plant Types field (on a Plant Asset).
    """

    def serialize(self, node, appstruct):
        if not appstruct:
            return []

        return [season.uuid.hex for season in appstruct]

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import SeasonRefsWidget

        model = self.app.model
        session = Session()

        if "values" not in kwargs:
            seasons = session.query(model.Season).order_by(model.Season.name).all()
            values = [(s.uuid.hex, str(s)) for s in seasons]
            kwargs["values"] = values

        return SeasonRefsWidget(self.request, **kwargs)


class StructureType(colander.SchemaType):

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null

        return json.dumps(appstruct)

    def widget_maker(self, **kwargs):  # pylint: disable=empty-docstring
        """ """
        from wuttafarm.web.forms.widgets import StructureWidget

        return StructureWidget(self.request, **kwargs)


class StructureTypeRef(ObjectRef):
    """
    Custom schema type for a
    :class:`~wuttafarm.db.model.structures.Structure` reference field.

    This is a subclass of
    :class:`~wuttaweb:wuttaweb.forms.schema.ObjectRef`.
    """

    @property
    def model_class(self):  # pylint: disable=empty-docstring
        """ """
        model = self.app.model
        return model.StructureType

    def sort_query(self, query):  # pylint: disable=empty-docstring
        """ """
        return query.order_by(self.model_class.name)

    def get_object_url(self, obj):  # pylint: disable=empty-docstring
        """ """
        structure_type = obj
        return self.request.route_url("structure_types.view", uuid=structure_type.uuid)


class UnitRef(ObjectRef):
    """
    Custom schema type for a :class:`~wuttafarm.db.model.units.Unit`
    reference field.

    This is a subclass of
    :class:`~wuttaweb:wuttaweb.forms.schema.ObjectRef`.
    """

    @property
    def model_class(self):
        model = self.app.model
        return model.Unit

    def sort_query(self, query):
        return query.order_by(self.model_class.name)

    def get_object_url(self, unit):
        return self.request.route_url("units.view", uuid=unit.uuid)


class UsersType(colander.SchemaType):

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null

        return json.dumps(appstruct)

    def widget_maker(self, **kwargs):  # pylint: disable=empty-docstring
        """ """
        from wuttafarm.web.forms.widgets import UsersWidget

        return UsersWidget(self.request, **kwargs)


class AssetRefs(WuttaSet):
    """
    Schema type for Assets field (on a Log record)
    """

    def __init__(
        self, request, for_asset=None, is_group=None, is_location=None, **kwargs
    ):
        super().__init__(request, **kwargs)
        self.is_group = is_group
        self.is_location = is_location
        self.for_asset = for_asset

    def serialize(self, node, appstruct):
        if not appstruct:
            return colander.null

        return {asset.uuid.hex for asset in appstruct}

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import AssetRefsWidget

        model = self.app.model
        session = Session()

        if "values" not in kwargs:
            query = session.query(model.Asset)
            if self.is_group is not None:
                query = query.join(model.GroupAsset)
            if self.is_location is not None:
                query = query.filter(model.Asset.is_location == self.is_location)
            if self.for_asset:
                query = query.filter(model.Asset.uuid != self.for_asset.uuid)
            query = query.order_by(model.Asset.asset_name)
            values = [(asset.uuid.hex, str(asset)) for asset in query]
            kwargs["values"] = values

        return AssetRefsWidget(self.request, **kwargs)


class QuantityRefs(colander.List):
    """
    Schema type for Quantities field (on a Log record)
    """

    def __init__(self, request):
        super().__init__()
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, node, appstruct):
        if not appstruct:
            return colander.null

        quantities = []
        for qty in appstruct:

            quantity = {
                "uuid": qty.uuid.hex,
                "quantity_type": {
                    "drupal_id": qty.quantity_type_id,
                    "name": qty.quantity_type.name,
                },
                "measure": qty.measure_id,
                "value": qty.get_value_decimal(),
                "units": {
                    "uuid": qty.units.uuid.hex,
                    "name": qty.units.name,
                },
                "as_text": qty.render_as_text(self.config),
                # nb. always include this regardless of quantity type,
                # for sake of easier frontend logic
                "material_types": [],
            }

            if qty.quantity_type_id == "material":
                quantity["material_types"] = []
                for mtype in qty.material_types:
                    quantity["material_types"].append(
                        {
                            "uuid": mtype.uuid.hex,
                            "name": mtype.name,
                        }
                    )

            quantities.append(quantity)

        return quantities

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import QuantityRefsWidget

        return QuantityRefsWidget(self.request, **kwargs)


class OwnerRefs(WuttaSet):
    """
    Schema type for Owners field (on a Log record)
    """

    def serialize(self, node, appstruct):
        if not appstruct:
            return colander.null

        return {user.uuid for user in appstruct}

    def widget_maker(self, **kwargs):
        from wuttafarm.web.forms.widgets import OwnerRefsWidget

        return OwnerRefsWidget(self.request, **kwargs)


class Notes(colander.String):
    """
    Custom schema type for "note" fields.
    """

    def serialize(self, node, appstruct):
        """ """
        if not appstruct:
            return colander.null

        return super().serialize(node, appstruct)

    def widget_maker(self, **kwargs):
        """
        Construct a default widget for the field.

        :returns: Instance of
           :class:`~wuttaweb.forms.widgets.NotesWidget`.
        """
        return NotesWidget(**kwargs)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaFarm Menu
"""

from wuttaweb import menus as base


class WuttaFarmMenuHandler(base.MenuHandler):
    """
    WuttaFarm menu handler
    """

    def make_menus(self, request, **kwargs):
        enum = self.app.enum
        mode = self.app.get_farmos_integration_mode()

        quick_menu = self.make_quick_menu(request)
        admin_menu = self.make_admin_menu(request, include_people=True)

        if mode == enum.FARMOS_INTEGRATION_MODE_WRAPPER:
            return [
                quick_menu,
                self.make_farmos_asset_menu(request),
                self.make_farmos_log_menu(request),
                self.make_farmos_other_menu(request),
                admin_menu,
            ]

        elif mode == enum.FARMOS_INTEGRATION_MODE_MIRROR:
            return [
                quick_menu,
                self.make_asset_menu(request),
                self.make_log_menu(request),
                self.make_farmos_full_menu(request),
                admin_menu,
            ]

        else:  # FARMOS_INTEGRATION_MODE_NONE
            return [
                quick_menu,
                self.make_asset_menu(request),
                self.make_log_menu(request),
                admin_menu,
            ]

    def make_quick_menu(self, request):
        return {
            "title": "Quick",
            "type": "menu",
            "items": [
                {
                    "title": "Eggs",
                    "route": "quick.eggs",
                    "perm": "quick.eggs",
                },
            ],
        }

    def make_asset_menu(self, request):
        return {
            "title": "Assets",
            "type": "menu",
            "items": [
                {
                    "title": "All Assets",
                    "route": "assets",
                    "perm": "assets.list",
                },
                {
                    "title": "Animal",
                    "route": "animal_assets",
                    "perm": "animal_assets.list",
                },
                {
                    "title": "Equipment",
                    "route": "equipment_assets",
                    "perm": "equipment_assets.list",
                },
                {
                    "title": "Group",
                    "route": "group_assets",
                    "perm": "group_assets.list",
                },
                {
                    "title": "Land",
                    "route": "land_assets",
                    "perm": "land_assets.list",
                },
                {
                    "title": "Plant",
                    "route": "plant_assets",
                    "perm": "plant_assets.list",
                },
                {
                    "title": "Structure",
                    "route": "structure_assets",
                    "perm": "structure_assets.list",
                },
                {
                    "title": "Water",
                    "route": "water_assets",
                    "perm": "water_assets.list",
                },
                {"type": "sep"},
                {
                    "title": "Animal Types",
                    "route": "animal_types",
                    "perm": "animal_types.list",
                },
                {
                    "title": "Equipment Types",
                    "route": "equipment_types",
                    "perm": "equipment_types.list",
                },
                {
                    "title": "Land Types",
                    "route": "land_types",
                    "perm": "land_types.list",
                },
                {
                    "title": "Plant Types",
                    "route": "plant_types",
                    "perm": "plant_types.list",
                },
                {
                    "title": "Seasons",
                    "route": "seasons",
                    "perm": "seasons.list",
                },
                {
                    "title": "Structure Types",
                    "route": "structure_types",
                    "perm": "structure_types.list",
                },
                {
                    "title": "Asset Types",
                    "route": "asset_types",
                    "perm": "asset_types.list",
                },
            ],
        }

    def make_log_menu(self, request):
        return {
            "title": "Logs",
            "type": "menu",
            "items": [
                {
                    "title": "All Logs",
                    "route": "log",
                    "perm": "log.list",
                },
                {
                    "title": "Activity",
                    "route": "logs_activity",
                    "perm": "logs_activity.list",
                },
                {
                    "title": "Harvest",
                    "route": "logs_harvest",
                    "perm": "logs_harvest.list",
                },
                {
                    "title": "Medical",
                    "route": "logs_medical",
                    "perm": "logs_medical.list",
                },
                {
                    "title": "Observation",
                    "route": "logs_observation",
                    "perm": "logs_observation.list",
                },
                {
                    "title": "Seeding",
                    "route": "logs_seeding",
                    "perm": "logs_seeding.list",
                },
                {"type": "sep"},
                {
                    "title": "All Quantities",
                    "route": "quantities",
                    "perm": "quantities.list",
                },
                {
                    "title": "Material Quantities",
                    "route": "quantities_material",
                    "perm": "quantities_material.list",
                },
                {
                    "title": "Standard Quantities",
                    "route": "quantities_standard",
                    "perm": "quantities_standard.list",
                },
                {"type": "sep"},
                {
                    "title": "Log Types",
                    "route": "log_types",
                    "perm": "log_types.list",
                },
                {
                    "title": "Material Types",
                    "route": "material_types",
                    "perm": "material_types.list",
                },
                {
                    "title": "Measures",
                    "route": "measures",
                    "perm": "measures.list",
                },
                {
                    "title": "Quantity Types",
                    "route": "quantity_types",
                    "perm": "quantity_types.list",
                },
                {
                    "title": "Units",
                    "route": "units",
                    "perm": "units.list",
                },
            ],
        }

    def make_farmos_full_menu(self, request):
        config = request.wutta_config
        app = config.get_app()
        return {
            "title": "farmOS",
            "type": "menu",
            "items": [
                {
                    "title": "Go to farmOS",
                    "url": app.get_farmos_url(),
                    "target": "_blank",
                },
                {"type": "sep"},
                {
                    "title": "Animal Assets",
                    "route": "farmos_animal_assets",
                    "perm": "farmos_animal_assets.list",
                },
                {
                    "title": "Equipment Assets",
                    "route": "farmos_equipment_assets",
                    "perm": "farmos_equipment_assets.list",
                },
                {
                    "title": "Group Assets",
                    "route": "farmos_group_assets",
                    "perm": "farmos_group_assets.list",
                },
                {
                    "title": "Land Assets",
                    "route": "farmos_land_assets",
                    "perm": "farmos_land_assets.list",
                },
                {
                    "title": "Plant Assets",
                    "route": "farmos_plant_assets",
                    "perm": "farmos_plant_assets.list",
                },
                {
                    "title": "Structure Assets",
                    "route": "farmos_structure_assets",
                    "perm": "farmos_structure_assets.list",
                },
                {
                    "title": "Water Assets",
                    "route": "farmos_water_assets",
                    "perm": "farmos_water_assets.list",
                },
                {"type": "sep"},
                {
                    "title": "Activity Logs",
                    "route": "farmos_logs_activity",
                    "perm": "farmos_logs_activity.list",
                },
                {
                    "title": "Harvest Logs",
                    "route": "farmos_logs_harvest",
                    "perm": "farmos_logs_harvest.list",
                },
                {
                    "title": "Medical Logs",
                    "route": "farmos_logs_medical",
                    "perm": "farmos_logs_medical.list",
                },
                {
                    "title": "Observation Logs",
                    "route": "farmos_logs_observation",
                    "perm": "farmos_logs_observation.list",
                },
                {
                    "title": "Seeding Logs",
                    "route": "farmos_logs_seeding",
                    "perm": "farmos_logs_seeding.list",
                },
                {"type": "sep"},
                {
                    "title": "Animal Types",
                    "route": "farmos_animal_types",
                    "perm": "farmos_animal_types.list",
                },
                {
                    "title": "Equipment Types",
                    "route": "farmos_equipment_types",
                    "perm": "farmos_equipment_types.list",
                },
                {
                    "title": "Land Types",
                    "route": "farmos_land_types",
                    "perm": "farmos_land_types.list",
                },
                {
                    "title": "Plant Types",
                    "route": "farmos_plant_types",
                    "perm": "farmos_plant_types.list",
                },
                {
                    "title": "Seasons",
                    "route": "farmos_seasons",
                    "perm": "farmos_seasons.list",
                },
                {
                    "title": "Structure Types",
                    "route": "farmos_structure_types",
                    "perm": "farmos_structure_types.list",
                },
                {"type": "sep"},
                {
                    "title": "Asset Types",
                    "route": "farmos_asset_types",
                    "perm": "farmos_asset_types.list",
                },
                {
                    "title": "Log Types",
                    "route": "farmos_log_types",
                    "perm": "farmos_log_types.list",
                },
                {
                    "title": "Material Types",
                    "route": "farmos_material_types",
                    "perm": "farmos_material_types.list",
                },
                {
                    "title": "Quantity Types",
                    "route": "farmos_quantity_types",
                    "perm": "farmos_quantity_types.list",
                },
                {
                    "title": "Material Quantities",
                    "route": "farmos_quantities_material",
                    "perm": "farmos_quantities_material.list",
                },
                {
                    "title": "Standard Quantities",
                    "route": "farmos_quantities_standard",
                    "perm": "farmos_quantities_standard.list",
                },
                {
                    "title": "Units",
                    "route": "farmos_units",
                    "perm": "farmos_units.list",
                },
                {"type": "sep"},
                {
                    "title": "Users",
                    "route": "farmos_users",
                    "perm": "farmos_users.list",
                },
            ],
        }

    def make_farmos_asset_menu(self, request):
        config = request.wutta_config
        app = config.get_app()
        return {
            "title": "Assets",
            "type": "menu",
            "items": [
                {
                    "title": "Animal",
                    "route": "farmos_animal_assets",
                    "perm": "farmos_animal_assets.list",
                },
                {
                    "title": "Equipment",
                    "route": "farmos_equipment_assets",
                    "perm": "farmos_equipment_assets.list",
                },
                {
                    "title": "Group",
                    "route": "farmos_group_assets",
                    "perm": "farmos_group_assets.list",
                },
                {
                    "title": "Land",
                    "route": "farmos_land_assets",
                    "perm": "farmos_land_assets.list",
                },
                {
                    "title": "Plant",
                    "route": "farmos_plant_assets",
                    "perm": "farmos_plant_assets.list",
                },
                {
                    "title": "Structure",
                    "route": "farmos_structure_assets",
                    "perm": "farmos_structure_assets.list",
                },
                {
                    "title": "Water",
                    "route": "farmos_water_assets",
                    "perm": "farmos_water_assets.list",
                },
                {"type": "sep"},
                {
                    "title": "Animal Types",
                    "route": "farmos_animal_types",
                    "perm": "farmos_animal_types.list",
                },
                {
                    "title": "Equipment Types",
                    "route": "farmos_equipment_types",
                    "perm": "farmos_equipment_types.list",
                },
                {
                    "title": "Land Types",
                    "route": "farmos_land_types",
                    "perm": "farmos_land_types.list",
                },
                {
                    "title": "Plant Types",
                    "route": "farmos_plant_types",
                    "perm": "farmos_plant_types.list",
                },
                {
                    "title": "Seasons",
                    "route": "farmos_seasons",
                    "perm": "farmos_seasons.list",
                },
                {
                    "title": "Structure Types",
                    "route": "farmos_structure_types",
                    "perm": "farmos_structure_types.list",
                },
                {"type": "sep"},
                {
                    "title": "Asset Types",
                    "route": "farmos_asset_types",
                    "perm": "farmos_asset_types.list",
                },
            ],
        }

    def make_farmos_log_menu(self, request):
        config = request.wutta_config
        app = config.get_app()
        return {
            "title": "Logs",
            "type": "menu",
            "items": [
                {
                    "title": "Activity",
                    "route": "farmos_logs_activity",
                    "perm": "farmos_logs_activity.list",
                },
                {
                    "title": "Harvest",
                    "route": "farmos_logs_harvest",
                    "perm": "farmos_logs_harvest.list",
                },
                {
                    "title": "Medical",
                    "route": "farmos_logs_medical",
                    "perm": "farmos_logs_medical.list",
                },
                {
                    "title": "Observation",
                    "route": "farmos_logs_observation",
                    "perm": "farmos_logs_observation.list",
                },
                {
                    "title": "Seeding",
                    "route": "farmos_logs_seeding",
                    "perm": "farmos_logs_seeding.list",
                },
                {"type": "sep"},
                {
                    "title": "Log Types",
                    "route": "farmos_log_types",
                    "perm": "farmos_log_types.list",
                },
                {
                    "title": "Material Types",
                    "route": "farmos_material_types",
                    "perm": "farmos_material_types.list",
                },
                {
                    "title": "Quantity Types",
                    "route": "farmos_quantity_types",
                    "perm": "farmos_quantity_types.list",
                },
                {
                    "title": "Material Quantities",
                    "route": "farmos_quantities_material",
                    "perm": "farmos_quantities_material.list",
                },
                {
                    "title": "Standard Quantities",
                    "route": "farmos_quantities_standard",
                    "perm": "farmos_quantities_standard.list",
                },
                {
                    "title": "Units",
                    "route": "farmos_units",
                    "perm": "farmos_units.list",
                },
            ],
        }

    def make_farmos_other_menu(self, request):
        config = request.wutta_config
        app = config.get_app()
        return {
            "title": "farmOS",
            "type": "menu",
            "items": [
                {
                    "title": "Go to farmOS",
                    "url": app.get_farmos_url(),
                    "target": "_blank",
                },
                {"type": "sep"},
                {
                    "title": "Users",
                    "route": "farmos_users",
                    "perm": "farmos_users.list",
                },
            ],
        }



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Asset handler
"""

from wuttjamaican.app import GenericHandler


class AssetHandler(GenericHandler):
    """
    Base class and default implementation for the asset
    :term:`handler`.
    """

    def get_groups(self, asset):
        model = self.app.model
        session = self.app.get_session(asset)

        grplog = (
            session.query(model.Log)
            .join(model.LogAsset)
            .filter(model.LogAsset.asset == asset)
            .filter(model.Log.is_group_assignment == True)
            .order_by(model.Log.timestamp.desc())
            .first()
        )
        if grplog:
            return grplog.groups
        return []

    def get_locations(self, asset):
        model = self.app.model
        session = self.app.get_session(asset)

        loclog = (
            session.query(model.Log)
            .join(model.LogAsset)
            .filter(model.LogAsset.asset == asset)
            .filter(model.Log.is_movement == True)
            .order_by(model.Log.timestamp.desc())
            .first()
        )
        if loclog:
            return loclog.locations
        return []



"""add Plant Assets and more Logs

Revision ID: 11e0e46f48a6
Revises: dd6351e69233
Create Date: 2026-02-18 18:11:46.536930

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "11e0e46f48a6"
down_revision: Union[str, None] = "dd6351e69233"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # plant_type
    op.create_table(
        "plant_type",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("name", sa.String(length=100), nullable=False),
        sa.Column("description", sa.String(length=255), nullable=True),
        sa.Column("farmos_uuid", wuttjamaican.db.util.UUID(), nullable=True),
        sa.Column("drupal_id", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_plant_type")),
        sa.UniqueConstraint("drupal_id", name=op.f("uq_plant_type_drupal_id")),
        sa.UniqueConstraint("farmos_uuid", name=op.f("uq_plant_type_farmos_uuid")),
        sa.UniqueConstraint("name", name=op.f("uq_plant_type_name")),
    )
    op.create_table(
        "plant_type_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column("name", sa.String(length=100), autoincrement=False, nullable=True),
        sa.Column(
            "description", sa.String(length=255), autoincrement=False, nullable=True
        ),
        sa.Column(
            "farmos_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("drupal_id", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_plant_type_version")
        ),
    )
    op.create_index(
        op.f("ix_plant_type_version_end_transaction_id"),
        "plant_type_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_plant_type_version_operation_type"),
        "plant_type_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_plant_type_version_pk_transaction_id",
        "plant_type_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_plant_type_version_pk_validity",
        "plant_type_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_plant_type_version_transaction_id"),
        "plant_type_version",
        ["transaction_id"],
        unique=False,
    )

    # asset_plant
    op.create_table(
        "asset_plant",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["uuid"], ["asset.uuid"], name=op.f("fk_asset_plant_uuid_asset")
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_asset_plant")),
    )
    op.create_table(
        "asset_plant_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_asset_plant_version")
        ),
    )
    op.create_index(
        op.f("ix_asset_plant_version_end_transaction_id"),
        "asset_plant_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_plant_version_operation_type"),
        "asset_plant_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_asset_plant_version_pk_transaction_id",
        "asset_plant_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_asset_plant_version_pk_validity",
        "asset_plant_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_plant_version_transaction_id"),
        "asset_plant_version",
        ["transaction_id"],
        unique=False,
    )

    # asset_plant_plant_type
    op.create_table(
        "asset_plant_plant_type",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("plant_asset_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("plant_type_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["plant_asset_uuid"],
            ["asset_plant.uuid"],
            name=op.f("fk_asset_plant_plant_type_plant_asset_uuid_asset_plant"),
        ),
        sa.ForeignKeyConstraint(
            ["plant_type_uuid"],
            ["plant_type.uuid"],
            name=op.f("fk_asset_plant_plant_type_plant_type_uuid_plant_type"),
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_asset_plant_plant_type")),
    )
    op.create_table(
        "asset_plant_plant_type_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "plant_asset_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "plant_type_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_asset_plant_plant_type_version")
        ),
    )
    op.create_index(
        op.f("ix_asset_plant_plant_type_version_end_transaction_id"),
        "asset_plant_plant_type_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_plant_plant_type_version_operation_type"),
        "asset_plant_plant_type_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_asset_plant_plant_type_version_pk_transaction_id",
        "asset_plant_plant_type_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_asset_plant_plant_type_version_pk_validity",
        "asset_plant_plant_type_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_asset_plant_plant_type_version_transaction_id"),
        "asset_plant_plant_type_version",
        ["transaction_id"],
        unique=False,
    )

    # log_asset
    op.create_table(
        "log_asset",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("log_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("asset_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["asset_uuid"], ["asset.uuid"], name=op.f("fk_log_asset_asset_uuid_asset")
        ),
        sa.ForeignKeyConstraint(
            ["log_uuid"], ["log.uuid"], name=op.f("fk_log_asset_log_uuid_log")
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_log_asset")),
    )
    op.create_table(
        "log_asset_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "log_uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "asset_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_log_asset_version")
        ),
    )
    op.create_index(
        op.f("ix_log_asset_version_end_transaction_id"),
        "log_asset_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_log_asset_version_operation_type"),
        "log_asset_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_log_asset_version_pk_transaction_id",
        "log_asset_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_log_asset_version_pk_validity",
        "log_asset_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_log_asset_version_transaction_id"),
        "log_asset_version",
        ["transaction_id"],
        unique=False,
    )

    # log_harvest
    op.create_table(
        "log_harvest",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["uuid"], ["log.uuid"], name=op.f("fk_log_harvest_uuid_log")
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_log_harvest")),
    )
    op.create_table(
        "log_harvest_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_log_harvest_version")
        ),
    )
    op.create_index(
        op.f("ix_log_harvest_version_end_transaction_id"),
        "log_harvest_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_log_harvest_version_operation_type"),
        "log_harvest_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_log_harvest_version_pk_transaction_id",
        "log_harvest_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_log_harvest_version_pk_validity",
        "log_harvest_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_log_harvest_version_transaction_id"),
        "log_harvest_version",
        ["transaction_id"],
        unique=False,
    )

    # log_medical
    op.create_table(
        "log_medical",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["uuid"], ["log.uuid"], name=op.f("fk_log_medical_uuid_log")
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_log_medical")),
    )
    op.create_table(
        "log_medical_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_log_medical_version")
        ),
    )
    op.create_index(
        op.f("ix_log_medical_version_end_transaction_id"),
        "log_medical_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_log_medical_version_operation_type"),
        "log_medical_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_log_medical_version_pk_transaction_id",
        "log_medical_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_log_medical_version_pk_validity",
        "log_medical_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_log_medical_version_transaction_id"),
        "log_medical_version",
        ["transaction_id"],
        unique=False,
    )

    # log_observation
    op.create_table(
        "log_observation",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["uuid"], ["log.uuid"], name=op.f("fk_log_observation_uuid_log")
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_log_observation")),
    )
    op.create_table(
        "log_observation_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_log_observation_version")
        ),
    )
    op.create_index(
        op.f("ix_log_observation_version_end_transaction_id"),
        "log_observation_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_log_observation_version_operation_type"),
        "log_observation_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_log_observation_version_pk_transaction_id",
        "log_observation_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_log_observation_version_pk_validity",
        "log_observation_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_log_observation_version_transaction_id"),
        "log_observation_version",
        ["transaction_id"],
        unique=False,
    )


def downgrade() -> None:

    # log_observation
    op.drop_index(
        op.f("ix_log_observation_version_transaction_id"),
        table_name="log_observation_version",
    )
    op.drop_index(
        "ix_log_observation_version_pk_validity", table_name="log_observation_version"
    )
    op.drop_index(
        "ix_log_observation_version_pk_transaction_id",
        table_name="log_observation_version",
    )
    op.drop_index(
        op.f("ix_log_observation_version_operation_type"),
        table_name="log_observation_version",
    )
    op.drop_index(
        op.f("ix_log_observation_version_end_transaction_id"),
        table_name="log_observation_version",
    )
    op.drop_table("log_observation_version")
    op.drop_table("log_observation")

    # log_medical
    op.drop_index(
        op.f("ix_log_medical_version_transaction_id"), table_name="log_medical_version"
    )
    op.drop_index(
        "ix_log_medical_version_pk_validity", table_name="log_medical_version"
    )
    op.drop_index(
        "ix_log_medical_version_pk_transaction_id", table_name="log_medical_version"
    )
    op.drop_index(
        op.f("ix_log_medical_version_operation_type"), table_name="log_medical_version"
    )
    op.drop_index(
        op.f("ix_log_medical_version_end_transaction_id"),
        table_name="log_medical_version",
    )
    op.drop_table("log_medical_version")
    op.drop_table("log_medical")

    # log_harvest
    op.drop_index(
        op.f("ix_log_harvest_version_transaction_id"), table_name="log_harvest_version"
    )
    op.drop_index(
        "ix_log_harvest_version_pk_validity", table_name="log_harvest_version"
    )
    op.drop_index(
        "ix_log_harvest_version_pk_transaction_id", table_name="log_harvest_version"
    )
    op.drop_index(
        op.f("ix_log_harvest_version_operation_type"), table_name="log_harvest_version"
    )
    op.drop_index(
        op.f("ix_log_harvest_version_end_transaction_id"),
        table_name="log_harvest_version",
    )
    op.drop_table("log_harvest_version")
    op.drop_table("log_harvest")

    # log_asset
    op.drop_index(
        op.f("ix_log_asset_version_transaction_id"), table_name="log_asset_version"
    )
    op.drop_index("ix_log_asset_version_pk_validity", table_name="log_asset_version")
    op.drop_index(
        "ix_log_asset_version_pk_transaction_id", table_name="log_asset_version"
    )
    op.drop_index(
        op.f("ix_log_asset_version_operation_type"), table_name="log_asset_version"
    )
    op.drop_index(
        op.f("ix_log_asset_version_end_transaction_id"), table_name="log_asset_version"
    )
    op.drop_table("log_asset_version")
    op.drop_table("log_asset")

    # asset_plant_plant_type
    op.drop_index(
        op.f("ix_asset_plant_plant_type_version_transaction_id"),
        table_name="asset_plant_plant_type_version",
    )
    op.drop_index(
        "ix_asset_plant_plant_type_version_pk_validity",
        table_name="asset_plant_plant_type_version",
    )
    op.drop_index(
        "ix_asset_plant_plant_type_version_pk_transaction_id",
        table_name="asset_plant_plant_type_version",
    )
    op.drop_index(
        op.f("ix_asset_plant_plant_type_version_operation_type"),
        table_name="asset_plant_plant_type_version",
    )
    op.drop_index(
        op.f("ix_asset_plant_plant_type_version_end_transaction_id"),
        table_name="asset_plant_plant_type_version",
    )
    op.drop_table("asset_plant_plant_type_version")
    op.drop_table("asset_plant_plant_type")

    # asset_plant
    op.drop_index(
        op.f("ix_asset_plant_version_transaction_id"), table_name="asset_plant_version"
    )
    op.drop_index(
        "ix_asset_plant_version_pk_validity", table_name="asset_plant_version"
    )
    op.drop_index(
        "ix_asset_plant_version_pk_transaction_id", table_name="asset_plant_version"
    )
    op.drop_index(
        op.f("ix_asset_plant_version_operation_type"), table_name="asset_plant_version"
    )
    op.drop_index(
        op.f("ix_asset_plant_version_end_transaction_id"),
        table_name="asset_plant_version",
    )
    op.drop_table("asset_plant_version")
    op.drop_table("asset_plant")

    # plant_type
    op.drop_index(
        op.f("ix_plant_type_version_transaction_id"), table_name="plant_type_version"
    )
    op.drop_index("ix_plant_type_version_pk_validity", table_name="plant_type_version")
    op.drop_index(
        "ix_plant_type_version_pk_transaction_id", table_name="plant_type_version"
    )
    op.drop_index(
        op.f("ix_plant_type_version_operation_type"), table_name="plant_type_version"
    )
    op.drop_index(
        op.f("ix_plant_type_version_end_transaction_id"),
        table_name="plant_type_version",
    )
    op.drop_table("plant_type_version")
    op.drop_table("plant_type")



"""convert active to archived

Revision ID: 8898184c5c75
Revises: 3e2ef02bf264
Create Date: 2026-02-14 18:41:23.042951

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "8898184c5c75"
down_revision: Union[str, None] = "3e2ef02bf264"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # animal
    op.alter_column("animal", "active", new_column_name="archived")
    animal = sa.sql.table(
        "animal",
        sa.sql.column("uuid"),
        sa.sql.column("archived"),
    )
    cursor = op.get_bind().execute(animal.select())
    for row in cursor.fetchall():
        op.get_bind().execute(
            animal.update()
            .where(animal.c.uuid == row.uuid)
            .values({"archived": not row.archived})
        )
    op.alter_column("animal_version", "active", new_column_name="archived")
    animal_version = sa.sql.table(
        "animal_version",
        sa.sql.column("uuid"),
        sa.sql.column("archived"),
    )
    cursor = op.get_bind().execute(animal_version.select())
    for row in cursor.fetchall():
        op.get_bind().execute(
            animal_version.update()
            .where(animal_version.c.uuid == row.uuid)
            .values({"archived": not row.archived})
        )

    # group
    op.alter_column("group", "active", new_column_name="archived")
    group = sa.sql.table(
        "group",
        sa.sql.column("uuid"),
        sa.sql.column("archived"),
    )
    cursor = op.get_bind().execute(group.select())
    for row in cursor.fetchall():
        op.get_bind().execute(
            group.update()
            .where(group.c.uuid == row.uuid)
            .values({"archived": not row.archived})
        )
    op.alter_column("group_version", "active", new_column_name="archived")
    group_version = sa.sql.table(
        "group_version",
        sa.sql.column("uuid"),
        sa.sql.column("archived"),
    )
    cursor = op.get_bind().execute(group_version.select())
    for row in cursor.fetchall():
        op.get_bind().execute(
            group_version.update()
            .where(group_version.c.uuid == row.uuid)
            .values({"archived": not row.archived})
        )

    # land_asset
    op.alter_column("land_asset", "active", new_column_name="archived")
    land_asset = sa.sql.table(
        "land_asset",
        sa.sql.column("uuid"),
        sa.sql.column("archived"),
    )
    cursor = op.get_bind().execute(land_asset.select())
    for row in cursor.fetchall():
        op.get_bind().execute(
            land_asset.update()
            .where(land_asset.c.uuid == row.uuid)
            .values({"archived": not row.archived})
        )
    op.alter_column("land_asset_version", "active", new_column_name="archived")
    land_asset_version = sa.sql.table(
        "land_asset_version",
        sa.sql.column("uuid"),
        sa.sql.column("archived"),
    )
    cursor = op.get_bind().execute(land_asset_version.select())
    for row in cursor.fetchall():
        op.get_bind().execute(
            land_asset_version.update()
            .where(land_asset_version.c.uuid == row.uuid)
            .values({"archived": not row.archived})
        )

    # structure
    op.alter_column("structure", "active", new_column_name="archived")
    structure = sa.sql.table(
        "structure",
        sa.sql.column("uuid"),
        sa.sql.column("archived"),
    )
    cursor = op.get_bind().execute(structure.select())
    for row in cursor.fetchall():
        op.get_bind().execute(
            structure.update()
            .where(structure.c.uuid == row.uuid)
            .values({"archived": not row.archived})
        )
    op.alter_column("structure_version", "active", new_column_name="archived")
    structure_version = sa.sql.table(
        "structure_version",
        sa.sql.column("uuid"),
        sa.sql.column("archived"),
    )
    cursor = op.get_bind().execute(structure_version.select())
    for row in cursor.fetchall():
        op.get_bind().execute(
            structure_version.update()
            .where(structure_version.c.uuid == row.uuid)
            .values({"archived": not row.archived})
        )


def downgrade() -> None:

    # structure
    op.alter_column("structure", "archived", new_column_name="active")
    structure = sa.sql.table(
        "structure",
        sa.sql.column("uuid"),
        sa.sql.column("active"),
    )
    cursor = op.get_bind().execute(structure.select())
    for row in cursor.fetchall():
        op.get_bind().execute(
            structure.update()
            .where(structure.c.uuid == row.uuid)
            .values({"active": not row.active})
        )
    op.alter_column("structure_version", "archived", new_column_name="active")
    structure_version = sa.sql.table(
        "structure_version",
        sa.sql.column("uuid"),
        sa.sql.column("active"),
    )
    cursor = op.get_bind().execute(structure_version.select())
    for row in cursor.fetchall():
        op.get_bind().execute(
            structure_version.update()
            .where(structure_version.c.uuid == row.uuid)
            .values({"active": not row.active})
        )

    # land_asset
    op.alter_column("land_asset", "archived", new_column_name="active")
    land_asset = sa.sql.table(
        "land_asset",
        sa.sql.column("uuid"),
        sa.sql.column("active"),
    )
    cursor = op.get_bind().execute(land_asset.select())
    for row in cursor.fetchall():
        op.get_bind().execute(
            land_asset.update()
            .where(land_asset.c.uuid == row.uuid)
            .values({"active": not row.active})
        )
    op.alter_column("land_asset_version", "archived", new_column_name="active")
    land_asset_version = sa.sql.table(
        "land_asset_version",
        sa.sql.column("uuid"),
        sa.sql.column("active"),
    )
    cursor = op.get_bind().execute(land_asset_version.select())
    for row in cursor.fetchall():
        op.get_bind().execute(
            land_asset_version.update()
            .where(land_asset_version.c.uuid == row.uuid)
            .values({"active": not row.active})
        )

    # group
    op.alter_column("group", "archived", new_column_name="active")
    group = sa.sql.table(
        "group",
        sa.sql.column("uuid"),
        sa.sql.column("active"),
    )
    cursor = op.get_bind().execute(group.select())
    for row in cursor.fetchall():
        op.get_bind().execute(
            group.update()
            .where(group.c.uuid == row.uuid)
            .values({"active": not row.active})
        )
    op.alter_column("group_version", "archived", new_column_name="active")
    group_version = sa.sql.table(
        "group_version",
        sa.sql.column("uuid"),
        sa.sql.column("active"),
    )
    cursor = op.get_bind().execute(group_version.select())
    for row in cursor.fetchall():
        op.get_bind().execute(
            group_version.update()
            .where(group_version.c.uuid == row.uuid)
            .values({"active": not row.active})
        )

    # animal
    op.alter_column("animal", "archived", new_column_name="active")
    animal = sa.sql.table(
        "animal",
        sa.sql.column("uuid"),
        sa.sql.column("active"),
    )
    cursor = op.get_bind().execute(animal.select())
    for row in cursor.fetchall():
        op.get_bind().execute(
            animal.update()
            .where(animal.c.uuid == row.uuid)
            .values({"active": not row.active})
        )
    op.alter_column("animal_version", "archived", new_column_name="active")
    animal_version = sa.sql.table(
        "animal_version",
        sa.sql.column("uuid"),
        sa.sql.column("active"),
    )
    cursor = op.get_bind().execute(animal_version.select())
    for row in cursor.fetchall():
        op.get_bind().execute(
            animal_version.update()
            .where(animal_version.c.uuid == row.uuid)
            .values({"active": not row.active})
        )



"""add MaterialQuantity

Revision ID: 9c53513f8862
Revises: 1c89f3fbb521
Create Date: 2026-03-08 18:14:05.587678

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "9c53513f8862"
down_revision: Union[str, None] = "1c89f3fbb521"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # quantity_material
    op.create_table(
        "quantity_material",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["uuid"], ["quantity.uuid"], name=op.f("fk_quantity_material_uuid_quantity")
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_quantity_material")),
    )
    op.create_table(
        "quantity_material_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_quantity_material_version")
        ),
    )
    op.create_index(
        op.f("ix_quantity_material_version_end_transaction_id"),
        "quantity_material_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_quantity_material_version_operation_type"),
        "quantity_material_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_quantity_material_version_pk_transaction_id",
        "quantity_material_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_quantity_material_version_pk_validity",
        "quantity_material_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_quantity_material_version_transaction_id"),
        "quantity_material_version",
        ["transaction_id"],
        unique=False,
    )

    # quantity_material_material_type
    op.create_table(
        "quantity_material_material_type",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("quantity_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("material_type_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["material_type_uuid"],
            ["material_type.uuid"],
            name=op.f(
                "fk_quantity_material_material_type_material_type_uuid_material_type"
            ),
        ),
        sa.ForeignKeyConstraint(
            ["quantity_uuid"],
            ["quantity_material.uuid"],
            name=op.f(
                "fk_quantity_material_material_type_quantity_uuid_quantity_material"
            ),
        ),
        sa.PrimaryKeyConstraint(
            "uuid", name=op.f("pk_quantity_material_material_type")
        ),
    )
    op.create_table(
        "quantity_material_material_type_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "quantity_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "material_type_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid",
            "transaction_id",
            name=op.f("pk_quantity_material_material_type_version"),
        ),
    )
    op.create_index(
        op.f("ix_quantity_material_material_type_version_end_transaction_id"),
        "quantity_material_material_type_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_quantity_material_material_type_version_operation_type"),
        "quantity_material_material_type_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        "ix_quantity_material_material_type_version_pk_transaction_id",
        "quantity_material_material_type_version",
        ["uuid", sa.literal_column("transaction_id DESC")],
        unique=False,
    )
    op.create_index(
        "ix_quantity_material_material_type_version_pk_validity",
        "quantity_material_material_type_version",
        ["uuid", "transaction_id", "end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_quantity_material_material_type_version_transaction_id"),
        "quantity_material_material_type_version",
        ["transaction_id"],
        unique=False,
    )


def downgrade() -> None:

    # quantity_material_material_type
    op.drop_index(
        op.f("ix_quantity_material_material_type_version_transaction_id"),
        table_name="quantity_material_material_type_version",
    )
    op.drop_index(
        "ix_quantity_material_material_type_version_pk_validity",
        table_name="quantity_material_material_type_version",
    )
    op.drop_index(
        "ix_quantity_material_material_type_version_pk_transaction_id",
        table_name="quantity_material_material_type_version",
    )
    op.drop_index(
        op.f("ix_quantity_material_material_type_version_operation_type"),
        table_name="quantity_material_material_type_version",
    )
    op.drop_index(
        op.f("ix_quantity_material_material_type_version_end_transaction_id"),
        table_name="quantity_material_material_type_version",
    )
    op.drop_table("quantity_material_material_type_version")
    op.drop_table("quantity_material_material_type")

    # quantity_material
    op.drop_index(
        op.f("ix_quantity_material_version_transaction_id"),
        table_name="quantity_material_version",
    )
    op.drop_index(
        "ix_quantity_material_version_pk_validity",
        table_name="quantity_material_version",
    )
    op.drop_index(
        "ix_quantity_material_version_pk_transaction_id",
        table_name="quantity_material_version",
    )
    op.drop_index(
        op.f("ix_quantity_material_version_operation_type"),
        table_name="quantity_material_version",
    )
    op.drop_index(
        op.f("ix_quantity_material_version_end_transaction_id"),
        table_name="quantity_material_version",
    )
    op.drop_table("quantity_material_version")
    op.drop_table("quantity_material")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Importer models targeting farmOS
"""

import datetime
from uuid import UUID

import requests

from wuttasync.importing import Importer


class ToFarmOS(Importer):
    """
    Base class for data importer targeting the farmOS API.
    """

    key = "uuid"
    caches_target = True

    def format_datetime(self, dt):
        """
        Convert a WuttaFarm datetime object to the format required for
        pushing to the farmOS API.
        """
        if dt is None:
            return None
        dt = self.app.localtime(dt)
        return dt.timestamp()

    def normalize_datetime(self, dt):
        """
        Convert a farmOS datetime value to naive UTC used by
        WuttaFarm.

        :param dt: Date/time string value "as-is" from the farmOS API.

        :returns: Equivalent naive UTC ``datetime``
        """
        if dt is None:
            return None
        dt = datetime.datetime.fromisoformat(dt)
        return self.app.make_utc(dt)


class ToFarmOSTaxonomy(ToFarmOS):

    farmos_taxonomy_type = None

    supported_fields = [
        "uuid",
        "name",
        "description",
    ]

    def get_target_objects(self, **kwargs):
        return list(
            self.farmos_client.resource.iterate(
                "taxonomy_term", self.farmos_taxonomy_type
            )
        )

    def get_target_object(self, key):

        # fetch from cache, if applicable
        if self.caches_target:
            return super().get_target_object(key)

        # okay now must fetch via API
        if self.get_keys() != ["uuid"]:
            raise ValueError("must use uuid key for this to work")
        uuid = key[0]

        try:
            result = self.farmos_client.resource.get_id(
                "taxonomy_term", self.farmos_taxonomy_type, str(uuid)
            )
        except requests.HTTPError as exc:
            if exc.response.status_code == 404:
                return None
            raise
        return result["data"]

    def normalize_target_object(self, obj):
        if description := obj["attributes"]["description"]:
            description = description["value"]
        return {
            "uuid": UUID(obj["id"]),
            "name": obj["attributes"]["name"],
            "description": description,
        }

    def get_term_payload(self, source_data):

        attrs = {}
        if "name" in self.fields:
            attrs["name"] = source_data["name"]
        if "description" in self.fields:
            attrs["description"] = {"value": source_data["description"]}

        payload = {"attributes": attrs}
        return payload

    def create_target_object(self, key, source_data):
        if source_data.get("__ignoreme__"):
            return None
        if self.dry_run:
            return source_data

        payload = self.get_term_payload(source_data)
        result = self.farmos_client.resource.send(
            "taxonomy_term", self.farmos_taxonomy_type, payload
        )
        normal = self.normalize_target_object(result["data"])
        normal["_new_object"] = result["data"]
        return normal

    def update_target_object(self, term, source_data, target_data=None):
        if self.dry_run:
            return term

        payload = self.get_term_payload(source_data)
        payload["id"] = str(source_data["uuid"])
        result = self.farmos_client.resource.send(
            "taxonomy_term", self.farmos_taxonomy_type, payload
        )
        return self.normalize_target_object(result["data"])


class ToFarmOSAsset(ToFarmOS):
    """
    Base class for asset data importer targeting the farmOS API.
    """

    farmos_asset_type = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.normal = self.app.get_normalizer(self.farmos_client)

    def get_target_objects(self, **kwargs):
        return list(self.farmos_client.asset.iterate(self.farmos_asset_type))

    def get_target_object(self, key):

        # fetch from cache, if applicable
        if self.caches_target:
            return super().get_target_object(key)

        # okay now must fetch via API
        if self.get_keys() != ["uuid"]:
            raise ValueError("must use uuid key for this to work")
        uuid = key[0]

        try:
            asset = self.farmos_client.asset.get_id(self.farmos_asset_type, str(uuid))
        except requests.HTTPError as exc:
            if exc.response.status_code == 404:
                return None
            raise
        return asset["data"]

    def create_target_object(self, key, source_data):
        if source_data.get("__ignoreme__"):
            return None
        if self.dry_run:
            return source_data

        payload = self.get_asset_payload(source_data)
        result = self.farmos_client.asset.send(self.farmos_asset_type, payload)
        normal = self.normalize_target_object(result["data"])
        normal["_new_object"] = result["data"]
        return normal

    def update_target_object(self, asset, source_data, target_data=None):
        if self.dry_run:
            return asset

        payload = self.get_asset_payload(source_data)
        payload["id"] = str(source_data["uuid"])
        result = self.farmos_client.asset.send(self.farmos_asset_type, payload)
        return self.normalize_target_object(result["data"])

    def normalize_target_object(self, asset):
        normal = self.normal.normalize_farmos_asset(asset)
        return {
            "uuid": UUID(normal["uuid"]),
            "asset_name": normal["asset_name"],
            "is_location": normal["is_location"],
            "is_fixed": normal["is_fixed"],
            # nb. this is only used for certain asset types
            "produces_eggs": normal["produces_eggs"],
            "parents": [(p["asset_type"], UUID(p["uuid"])) for p in normal["parents"]],
            "notes": normal["notes"],
            "archived": normal["archived"],
        }

    def get_asset_payload(self, source_data):

        attrs = {}
        if "asset_name" in self.fields:
            attrs["name"] = source_data["asset_name"]
        if "is_location" in self.fields:
            attrs["is_location"] = source_data["is_location"]
        if "is_fixed" in self.fields:
            attrs["is_fixed"] = source_data["is_fixed"]
        if "produces_eggs" in self.fields:
            attrs["produces_eggs"] = source_data["produces_eggs"]
        if "notes" in self.fields:
            attrs["notes"] = {"value": source_data["notes"]}
        if "archived" in self.fields:
            attrs["archived"] = source_data["archived"]

        rels = {}
        if "parents" in self.fields:
            rels["parent"] = {"data": []}
            for asset_type, uuid in source_data["parents"]:
                rels["parent"]["data"].append(
                    {
                        "id": str(uuid),
                        "type": f"asset--{asset_type}",
                    }
                )

        payload = {"attributes": attrs, "relationships": rels}
        return payload


class UnitImporter(ToFarmOSTaxonomy):

    model_title = "Unit"
    farmos_taxonomy_type = "unit"


class AnimalAssetImporter(ToFarmOSAsset):

    model_title = "AnimalAsset"
    farmos_asset_type = "animal"

    supported_fields = [
        "uuid",
        "asset_name",
        "animal_type_uuid",
        "sex",
        "is_sterile",
        "produces_eggs",
        "birthdate",
        "is_location",
        "is_fixed",
        "notes",
        "archived",
    ]

    def normalize_target_object(self, animal):
        data = super().normalize_target_object(animal)
        data.update(
            {
                "animal_type_uuid": UUID(
                    animal["relationships"]["animal_type"]["data"]["id"]
                ),
                "sex": animal["attributes"]["sex"],
                "is_sterile": animal["attributes"]["is_sterile"],
                "birthdate": self.normalize_datetime(animal["attributes"]["birthdate"]),
            }
        )
        return data

    def get_asset_payload(self, source_data):
        payload = super().get_asset_payload(source_data)

        attrs = {}
        if "sex" in self.fields:
            attrs["sex"] = source_data["sex"]
        if "is_sterile" in self.fields:
            attrs["is_sterile"] = source_data["is_sterile"]
        if "birthdate" in self.fields:
            attrs["birthdate"] = self.format_datetime(source_data["birthdate"])

        rels = {}
        if "animal_type_uuid" in self.fields:
            rels["animal_type"] = {
                "data": {
                    "id": str(source_data["animal_type_uuid"]),
                    "type": "taxonomy_term--animal_type",
                }
            }

        payload["attributes"].update(attrs)
        if rels:
            payload.setdefault("relationships", {}).update(rels)

        return payload


class AnimalTypeImporter(ToFarmOSTaxonomy):

    model_title = "AnimalType"
    farmos_taxonomy_type = "animal_type"


class EquipmentTypeImporter(ToFarmOSTaxonomy):

    model_title = "EquipmentType"
    farmos_taxonomy_type = "equipment_type"


class EquipmentAssetImporter(ToFarmOSAsset):

    model_title = "EquipmentAsset"
    farmos_asset_type = "equipment"

    supported_fields = [
        "uuid",
        "asset_name",
        "manufacturer",
        "model",
        "serial_number",
        "equipment_type_uuids",
        "is_location",
        "is_fixed",
        "notes",
        "archived",
    ]

    def normalize_target_object(self, equipment):
        data = super().normalize_target_object(equipment)
        data.update(
            {
                "manufacturer": equipment["attributes"]["manufacturer"],
                "model": equipment["attributes"]["model"],
                "serial_number": equipment["attributes"]["serial_number"],
                "equipment_type_uuids": [
                    UUID(etype["id"])
                    for etype in equipment["relationships"]["equipment_type"]["data"]
                ],
            }
        )
        return data

    def get_asset_payload(self, source_data):
        payload = super().get_asset_payload(source_data)

        attrs = {}
        if "manufacturer" in self.fields:
            attrs["manufacturer"] = source_data["manufacturer"]
        if "model" in self.fields:
            attrs["model"] = source_data["model"]
        if "serial_number" in self.fields:
            attrs["serial_number"] = source_data["serial_number"]

        rels = {}
        if "equipment_type_uuids" in self.fields:
            rels["equipment_type"] = {"data": []}
            for uuid in source_data["equipment_type_uuids"]:
                rels["equipment_type"]["data"].append(
                    {
                        "id": str(uuid),
                        "type": "taxonomy_term--equipment_type",
                    }
                )

        payload["attributes"].update(attrs)
        if rels:
            payload.setdefault("relationships", {}).update(rels)

        return payload


class MaterialTypeImporter(ToFarmOSTaxonomy):

    model_title = "MaterialType"
    farmos_taxonomy_type = "material_type"


class GroupAssetImporter(ToFarmOSAsset):

    model_title = "GroupAsset"
    farmos_asset_type = "group"

    supported_fields = [
        "uuid",
        "asset_name",
        "produces_eggs",
        "notes",
        "archived",
    ]


class LandAssetImporter(ToFarmOSAsset):

    model_title = "LandAsset"
    farmos_asset_type = "land"

    supported_fields = [
        "uuid",
        "asset_name",
        "land_type_id",
        "is_location",
        "is_fixed",
        "notes",
        "archived",
    ]

    def normalize_target_object(self, land):
        data = super().normalize_target_object(land)
        data.update(
            {
                "land_type_id": land["attributes"]["land_type"],
            }
        )
        return data

    def get_asset_payload(self, source_data):
        payload = super().get_asset_payload(source_data)

        attrs = {}
        if "land_type_id" in self.fields:
            attrs["land_type"] = source_data["land_type_id"]

        if attrs:
            payload["attributes"].update(attrs)

        return payload


class PlantTypeImporter(ToFarmOSTaxonomy):

    model_title = "PlantType"
    farmos_taxonomy_type = "plant_type"


class SeasonImporter(ToFarmOSTaxonomy):

    model_title = "Season"
    farmos_taxonomy_type = "season"


class PlantAssetImporter(ToFarmOSAsset):

    model_title = "PlantAsset"
    farmos_asset_type = "plant"

    supported_fields = [
        "uuid",
        "asset_name",
        "plant_type_uuids",
        "season_uuids",
        "notes",
        "archived",
    ]

    def normalize_target_object(self, plant):
        data = super().normalize_target_object(plant)
        data.update(
            {
                "plant_type_uuids": [
                    UUID(p["id"]) for p in plant["relationships"]["plant_type"]["data"]
                ],
                "season_uuids": [
                    UUID(p["id"]) for p in plant["relationships"]["season"]["data"]
                ],
            }
        )
        return data

    def get_asset_payload(self, source_data):
        payload = super().get_asset_payload(source_data)

        attrs = {}
        if "sex" in self.fields:
            attrs["sex"] = source_data["sex"]
        if "is_sterile" in self.fields:
            attrs["is_sterile"] = source_data["is_sterile"]
        if "birthdate" in self.fields:
            attrs["birthdate"] = self.format_datetime(source_data["birthdate"])

        rels = {}
        if "plant_type_uuids" in self.fields:
            rels["plant_type"] = {"data": []}
            for uuid in source_data["plant_type_uuids"]:
                rels["plant_type"]["data"].append(
                    {
                        "id": str(uuid),
                        "type": "taxonomy_term--plant_type",
                    }
                )
        if "season_uuids" in self.fields:
            rels["season"] = {"data": []}
            for uuid in source_data["season_uuids"]:
                rels["season"]["data"].append(
                    {
                        "id": str(uuid),
                        "type": "taxonomy_term--season",
                    }
                )

        payload["attributes"].update(attrs)
        if rels:
            payload.setdefault("relationships", {}).update(rels)

        return payload


class StructureAssetImporter(ToFarmOSAsset):

    model_title = "StructureAsset"
    farmos_asset_type = "structure"

    supported_fields = [
        "uuid",
        "asset_name",
        "structure_type_id",
        "is_location",
        "is_fixed",
        "notes",
        "archived",
    ]

    def normalize_target_object(self, structure):
        data = super().normalize_target_object(structure)
        data.update(
            {
                "structure_type_id": structure["attributes"]["structure_type"],
            }
        )
        return data

    def get_asset_payload(self, source_data):
        payload = super().get_asset_payload(source_data)

        attrs = {}
        if "structure_type_id" in self.fields:
            attrs["structure_type"] = source_data["structure_type_id"]

        if attrs:
            payload["attributes"].update(attrs)

        return payload


class WaterAssetImporter(ToFarmOSAsset):

    model_title = "WaterAsset"
    farmos_asset_type = "water"

    supported_fields = [
        "uuid",
        "asset_name",
        "is_location",
        "is_fixed",
        "notes",
        "archived",
    ]


##############################
# quantity importers
##############################


class ToFarmOSQuantity(ToFarmOS):
    """
    Base class for quantity data importer targeting the farmOS API.
    """

    farmos_quantity_type = None

    supported_fields = [
        "uuid",
        "measure",
        "value_numerator",
        "value_denominator",
        "label",
        "quantity_type_uuid",
        "unit_uuid",
    ]

    def get_target_objects(self, **kwargs):
        return list(
            self.farmos_client.resource.iterate("quantity", self.farmos_quantity_type)
        )

    def get_target_object(self, key):

        # fetch from cache, if applicable
        if self.caches_target:
            return super().get_target_object(key)

        # okay now must fetch via API
        if self.get_keys() != ["uuid"]:
            raise ValueError("must use uuid key for this to work")
        uuid = key[0]

        try:
            qty = self.farmos_client.resource.get_id(
                "quantity", self.farmos_quantity_type, str(uuid)
            )
        except requests.HTTPError as exc:
            if exc.response.status_code == 404:
                return None
            raise
        return qty["data"]

    def create_target_object(self, key, source_data):
        if source_data.get("__ignoreme__"):
            return None
        if self.dry_run:
            return source_data

        payload = self.get_quantity_payload(source_data)
        result = self.farmos_client.resource.send(
            "quantity", self.farmos_quantity_type, payload
        )
        normal = self.normalize_target_object(result["data"])
        normal["_new_object"] = result["data"]
        return normal

    def update_target_object(self, quantity, source_data, target_data=None):
        if self.dry_run:
            return quantity

        payload = self.get_quantity_payload(source_data)
        payload["id"] = str(source_data["uuid"])
        result = self.farmos_client.resource.send(
            "quantity", self.farmos_quantity_type, payload
        )
        return self.normalize_target_object(result["data"])

    def normalize_target_object(self, qty):

        result = {
            "uuid": UUID(qty["id"]),
            "measure": qty["attributes"]["measure"],
            "value_numerator": qty["attributes"]["value"]["numerator"],
            "value_denominator": qty["attributes"]["value"]["denominator"],
            "label": qty["attributes"]["label"],
            "quantity_type_uuid": UUID(
                qty["relationships"]["quantity_type"]["data"]["id"]
            ),
            "unit_uuid": None,
        }

        if unit := qty["relationships"]["units"]["data"]:
            result["unit_uuid"] = UUID(unit["id"])

        return result

    def get_quantity_payload(self, source_data):

        attrs = {}
        if "measure" in self.fields:
            attrs["measure"] = source_data["measure"]
        if "value_numerator" in self.fields and "value_denominator" in self.fields:
            attrs["value"] = {
                "numerator": source_data["value_numerator"],
                "denominator": source_data["value_denominator"],
            }
        if "label" in self.fields:
            attrs["label"] = source_data["label"]

        rels = {}
        if "quantity_type_uuid" in self.fields:
            rels["quantity_type"] = {
                "data": {
                    "id": str(source_data["quantity_type_uuid"]),
                    "type": "quantity_type--quantity_type",
                }
            }
        if "unit_uuid" in self.fields:
            rels["units"] = {
                "data": {
                    "id": str(source_data["unit_uuid"]),
                    "type": "taxonomy_term--unit",
                }
            }

        payload = {"attributes": attrs, "relationships": rels}

        return payload


class MaterialQuantityImporter(ToFarmOSQuantity):

    model_title = "MaterialQuantity"
    farmos_quantity_type = "material"

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "material_types",
            ]
        )
        return fields

    def normalize_target_object(self, quantity):
        data = super().normalize_target_object(quantity)

        if "material_types" in self.fields:
            data["material_types"] = [
                UUID(mtype["id"])
                for mtype in quantity["relationships"]["material_type"]["data"]
            ]

        return data

    def get_quantity_payload(self, source_data):
        payload = super().get_quantity_payload(source_data)

        rels = {}
        if "material_types" in self.fields:
            rels["material_type"] = {"data": []}
            for uuid in source_data["material_types"]:
                rels["material_type"]["data"].append(
                    {
                        "id": str(uuid),
                        "type": "taxonomy_term--material_type",
                    }
                )

        payload.setdefault("relationships", {}).update(rels)
        return payload


class StandardQuantityImporter(ToFarmOSQuantity):

    model_title = "StandardQuantity"
    farmos_quantity_type = "standard"


##############################
# log importers
##############################


class ToFarmOSLog(ToFarmOS):
    """
    Base class for log data importer targeting the farmOS API.
    """

    farmos_log_type = None

    supported_fields = [
        "uuid",
        "name",
        "timestamp",
        "is_movement",
        "is_group_assignment",
        "status",
        "notes",
        "quick",
        "assets",
        "locations",
        "groups",
        "quantities",
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.normal = self.app.get_normalizer(self.farmos_client)

    def get_target_objects(self, **kwargs):
        return list(self.farmos_client.log.iterate(self.farmos_log_type))

    def get_target_object(self, key):

        # fetch from cache, if applicable
        if self.caches_target:
            return super().get_target_object(key)

        # okay now must fetch via API
        if self.get_keys() != ["uuid"]:
            raise ValueError("must use uuid key for this to work")
        uuid = key[0]

        try:
            log = self.farmos_client.log.get_id(self.farmos_log_type, str(uuid))
        except requests.HTTPError as exc:
            if exc.response.status_code == 404:
                return None
            raise
        return log["data"]

    def create_target_object(self, key, source_data):
        if source_data.get("__ignoreme__"):
            return None
        if self.dry_run:
            return source_data

        payload = self.get_log_payload(source_data)
        result = self.farmos_client.log.send(self.farmos_log_type, payload)
        normal = self.normalize_target_object(result["data"])
        normal["_new_object"] = result["data"]
        return normal

    def update_target_object(self, asset, source_data, target_data=None):
        if self.dry_run:
            return asset

        payload = self.get_log_payload(source_data)
        payload["id"] = str(source_data["uuid"])
        result = self.farmos_client.log.send(self.farmos_log_type, payload)
        return self.normalize_target_object(result["data"])

    def normalize_target_object(self, log):
        normal = self.normal.normalize_farmos_log(log)
        return {
            "uuid": UUID(normal["uuid"]),
            "name": normal["name"],
            "timestamp": self.app.make_utc(normal["timestamp"]),
            "is_movement": normal["is_movement"],
            "is_group_assignment": normal["is_group_assignment"],
            "status": normal["status"],
            "notes": normal["notes"],
            "quick": normal["quick"],
            "assets": [(a["asset_type"], UUID(a["uuid"])) for a in normal["assets"]],
            "locations": [
                (l["asset_type"], UUID(l["uuid"])) for l in normal["locations"]
            ],
            "groups": [(g["asset_type"], UUID(g["uuid"])) for g in normal["groups"]],
            "quantities": [UUID(uuid) for uuid in normal["quantity_uuids"]],
        }

    def get_log_payload(self, source_data):

        attrs = {}
        if "name" in self.fields:
            attrs["name"] = source_data["name"]
        if "timestamp" in self.fields:
            attrs["timestamp"] = self.format_datetime(source_data["timestamp"])
        if "is_movement" in self.fields:
            attrs["is_movement"] = source_data["is_movement"]
        if "is_group_assignment" in self.fields:
            attrs["is_group_assignment"] = source_data["is_group_assignment"]
        if "status" in self.fields:
            attrs["status"] = source_data["status"]
        if "notes" in self.fields:
            attrs["notes"] = {"value": source_data["notes"]}
        if "quick" in self.fields:
            attrs["quick"] = source_data["quick"]

        rels = {}
        if "assets" in self.fields:
            assets = []
            for asset_type, uuid in source_data["assets"]:
                assets.append(
                    {
                        "type": f"asset--{asset_type}",
                        "id": str(uuid),
                    }
                )
            rels["asset"] = {"data": assets}
        if "locations" in self.fields:
            locations = []
            for asset_type, uuid in source_data["locations"]:
                locations.append(
                    {
                        "type": f"asset--{asset_type}",
                        "id": str(uuid),
                    }
                )
            rels["location"] = {"data": locations}
        if "groups" in self.fields:
            groups = []
            for asset_type, uuid in source_data["groups"]:
                groups.append(
                    {
                        "type": f"asset--{asset_type}",
                        "id": str(uuid),
                    }
                )
            rels["group"] = {"data": groups}
        if "quantities" in self.fields:
            quantities = []
            for uuid in source_data["quantities"]:
                quantities.append(
                    {
                        # TODO: support other quantity types
                        "type": "quantity--standard",
                        "id": str(uuid),
                    }
                )
            rels["quantity"] = {"data": quantities}

        payload = {"attributes": attrs, "relationships": rels}
        return payload


class ActivityLogImporter(ToFarmOSLog):

    model_title = "ActivityLog"
    farmos_log_type = "activity"


class HarvestLogImporter(ToFarmOSLog):

    model_title = "HarvestLog"
    farmos_log_type = "harvest"


class MedicalLogImporter(ToFarmOSLog):

    model_title = "MedicalLog"
    farmos_log_type = "medical"

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "vet",
            ]
        )
        return fields

    def normalize_target_object(self, log):
        data = super().normalize_target_object(log)
        data.update(
            {
                "vet": log["attributes"]["vet"],
            }
        )
        return data

    def get_log_payload(self, source_data):
        payload = super().get_log_payload(source_data)

        if "vet" in self.fields:
            payload["attributes"]["vet"] = source_data["vet"]

        return payload


class ObservationLogImporter(ToFarmOSLog):

    model_title = "ObservationLog"
    farmos_log_type = "observation"


class SeedingLogImporter(ToFarmOSLog):

    model_title = "SeedingLog"
    farmos_log_type = "seeding"

    def get_supported_fields(self):
        fields = list(super().get_supported_fields())
        fields.extend(
            [
                "source",
                "purchase_date",
                "lot_number",
            ]
        )
        return fields

    def normalize_target_object(self, log):
        data = super().normalize_target_object(log)
        data.update(
            {
                "source": log["attributes"]["source"],
                "purchase_date": self.normalize_datetime(
                    log["attributes"]["purchase_date"]
                ),
                "lot_number": log["attributes"]["lot_number"],
            }
        )
        return data

    def get_log_payload(self, source_data):
        payload = super().get_log_payload(source_data)

        attrs = {}
        if "source" in self.fields:
            attrs["source"] = source_data["source"]
        if "purchase_date" in self.fields:
            attrs["purchase_date"] = self.format_datetime(source_data["purchase_date"])
        if "lot_number" in self.fields:
            attrs["lot_number"] = source_data["lot_number"]

        if attrs:
            payload["attributes"].update(attrs)
        return payload



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Units
"""

from wuttafarm.web.views import WuttaFarmMasterView
from wuttafarm.db.model import Measure, Unit, Quantity


class MeasureView(WuttaFarmMasterView):
    """
    Master view for Measures
    """

    model_class = Measure
    route_prefix = "measures"
    url_prefix = "/measures"

    grid_columns = [
        "ordinal",
        "name",
        "drupal_id",
    ]

    sort_defaults = "ordinal"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "ordinal",
        "name",
        "drupal_id",
    ]

    has_rows = True
    row_model_class = Quantity
    rows_viewable = True

    row_labels = {
        "quantity_type_id": "Quantity Type ID",
        "measure_id": "Measure ID",
    }

    row_grid_columns = [
        "drupal_id",
        "as_text",
        "quantity_type",
        "value",
        "units",
        "label",
    ]

    rows_sort_defaults = ("drupal_id", "desc")

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")

    def get_row_grid_data(self, measure):
        model = self.app.model
        session = self.Session()
        return session.query(model.Quantity).filter(model.Quantity.measure == measure)

    def configure_row_grid(self, grid):
        g = grid
        super().configure_row_grid(g)

        # drupal_id
        g.set_label("drupal_id", "ID", column_only=True)

        # as_text
        g.set_renderer("as_text", self.render_as_text_for_grid)
        g.set_link("as_text")

        # value
        g.set_renderer("value", self.render_value_for_grid)

    def render_as_text_for_grid(self, quantity, field, value):
        return quantity.render_as_text(self.config)

    def render_value_for_grid(self, quantity, field, value):
        value = quantity.value_numerator / quantity.value_denominator
        return self.app.render_quantity(value)

    def get_row_action_url_view(self, quantity, i):
        return self.request.route_url(
            f"quantities_{quantity.quantity_type_id}.view", uuid=quantity.uuid
        )

    @classmethod
    def defaults(cls, config):
        """ """
        wutta_config = config.registry.settings.get("wutta_config")
        app = wutta_config.get_app()

        if app.is_farmos_mirror():
            cls.creatable = False
            cls.editable = False
            cls.deletable = False

        cls._defaults(config)


class UnitView(WuttaFarmMasterView):
    """
    Master view for Units
    """

    model_class = Unit
    route_prefix = "units"
    url_prefix = "/units"

    farmos_entity_type = "taxonomy_term"
    farmos_bundle = "unit"
    farmos_refurl_path = "/admin/structure/taxonomy/manage/unit/overview"

    grid_columns = [
        "name",
        "description",
    ]

    sort_defaults = "name"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "name",
        "description",
        "drupal_id",
        "farmos_uuid",
    ]

    has_rows = True
    row_model_class = Quantity
    rows_viewable = True

    row_labels = {
        "quantity_type_id": "Quantity Type ID",
        "measure_id": "Measure ID",
    }

    row_grid_columns = [
        "drupal_id",
        "as_text",
        "quantity_type",
        "measure",
        "value",
        "label",
    ]

    rows_sort_defaults = ("drupal_id", "desc")

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")

    def get_farmos_url(self, unit):
        return self.app.get_farmos_url(f"/taxonomy/term/{unit.drupal_id}")

    def get_xref_buttons(self, unit):
        buttons = super().get_xref_buttons(unit)

        if unit.farmos_uuid:
            buttons.append(
                self.make_button(
                    "View farmOS record",
                    primary=True,
                    url=self.request.route_url(
                        "farmos_units.view", uuid=unit.farmos_uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons

    def get_row_grid_data(self, unit):
        model = self.app.model
        session = self.Session()
        return session.query(model.Quantity).filter(model.Quantity.units == unit)

    def configure_row_grid(self, grid):
        g = grid
        super().configure_row_grid(g)

        # drupal_id
        g.set_label("drupal_id", "ID", column_only=True)

        # as_text
        g.set_renderer("as_text", self.render_as_text_for_grid)
        g.set_link("as_text")

        # value
        g.set_renderer("value", self.render_value_for_grid)

    def render_as_text_for_grid(self, quantity, field, value):
        return quantity.render_as_text(self.config)

    def render_value_for_grid(self, quantity, field, value):
        value = quantity.value_numerator / quantity.value_denominator
        return self.app.render_quantity(value)

    def get_row_action_url_view(self, quantity, i):
        return self.request.route_url(
            f"quantities_{quantity.quantity_type_id}.view", uuid=quantity.uuid
        )


def defaults(config, **kwargs):
    base = globals()

    MeasureView = kwargs.get("MeasureView", base["MeasureView"])
    MeasureView.defaults(config)

    UnitView = kwargs.get("UnitView", base["UnitView"])
    UnitView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base views for Logs
"""

from collections import OrderedDict

import colander
from webhelpers2.html import tags, HTML

from wuttaweb.forms.schema import WuttaDictEnum
from wuttaweb.db import Session
from wuttaweb.forms.widgets import WuttaDateTimeWidget

from wuttafarm.web.views import WuttaFarmMasterView
from wuttafarm.db.model import LogType, Log
from wuttafarm.web.forms.schema import AssetRefs, QuantityRefs, OwnerRefs
from wuttafarm.util import get_log_type_enum


class LogTypeView(WuttaFarmMasterView):
    """
    Master view for Log Types
    """

    model_class = LogType
    route_prefix = "log_types"
    url_prefix = "/log-types"

    grid_columns = [
        "name",
        "description",
    ]

    sort_defaults = "name"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "name",
        "description",
        "drupal_id",
        "farmos_uuid",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")

    def get_xref_buttons(self, log_type):
        buttons = super().get_xref_buttons(log_type)

        if log_type.farmos_uuid:
            buttons.append(
                self.make_button(
                    "View farmOS record",
                    primary=True,
                    url=self.request.route_url(
                        "farmos_log_types.view", uuid=log_type.farmos_uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


class LogMasterView(WuttaFarmMasterView):
    """
    Base class for Asset master views
    """

    farmos_entity_type = "log"

    labels = {
        "message": "Log Name",
        "locations": "Location",
        "quantities": "Quantity",
    }

    grid_columns = [
        "status",
        "drupal_id",
        "timestamp",
        "message",
        "assets",
        "locations",
        "quantities",
        "is_group_assignment",
        "owners",
    ]

    sort_defaults = ("timestamp", "desc")

    filter_defaults = {
        "message": {"active": True, "verb": "contains"},
        "status": {"active": True, "verb": "not_equal", "value": "abandoned"},
    }

    form_fields = [
        "message",
        "timestamp",
        "assets",
        "groups",
        "locations",
        "quantities",
        "notes",
        "status",
        "log_type",
        "owners",
        "is_movement",
        "is_group_assignment",
        "quick",
        "drupal_id",
        "farmos_uuid",
    ]

    def get_query(self, session=None):
        """ """
        model = self.app.model
        model_class = self.get_model_class()
        session = session or self.Session()
        query = session.query(model_class)
        if model_class is not model.Log:
            query = query.join(model.Log)
        return query

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)
        model = self.app.model
        enum = self.app.enum

        # status
        g.set_enum("status", enum.LOG_STATUS)
        g.set_sorter("status", model.Log.status)
        g.set_filter(
            "status",
            model.Log.status,
            verbs=["equal", "not_equal"],
            choices=enum.LOG_STATUS,
        )

        # drupal_id
        g.set_label("drupal_id", "ID", column_only=True)
        g.set_sorter("drupal_id", model.Log.drupal_id)
        g.set_filter("drupal_id", model.Log.drupal_id)

        # timestamp
        g.set_renderer("timestamp", "date")
        g.set_link("timestamp")
        g.set_sorter("timestamp", model.Log.timestamp)
        g.set_filter("timestamp", model.Log.timestamp)

        # message
        g.set_link("message")
        g.set_sorter("message", model.Log.message)
        g.set_filter("message", model.Log.message)

        # assets
        g.set_renderer("assets", self.render_assets_for_grid)

        # groups
        g.set_renderer("groups", self.render_assets_for_grid)

        # locations
        g.set_renderer("locations", self.render_assets_for_grid)

        # quantities
        g.set_renderer("quantities", self.render_quantities_for_grid)

        # is_group_assignment
        g.set_renderer("is_group_assignment", "boolean")
        g.set_sorter("is_group_assignment", model.Log.is_group_assignment)
        g.set_filter("is_group_assignment", model.Log.is_group_assignment)

        # owners
        g.set_label("owners", "Owner")
        g.set_renderer("owners", self.render_owners_for_grid)

    def render_assets_for_grid(self, log, field, value):
        assets = getattr(log, field)

        if self.farmos_style_grid_links:
            links = []
            for asset in assets:
                url = self.request.route_url(
                    f"{asset.asset_type}_assets.view", uuid=asset.uuid
                )
                links.append(tags.link_to(str(asset), url))
            return ", ".join(links)

        return ", ".join([str(a) for a in assets])

    def render_quantities_for_grid(self, log, field, value):
        quantities = getattr(log, field) or []
        items = []
        for qty in quantities:
            items.append(HTML.tag("li", c=qty.render_as_text(self.config)))
        return HTML.tag("ul", c=items)

    def render_owners_for_grid(self, log, field, value):

        if self.farmos_style_grid_links:
            links = []
            for user in log.owners:
                url = self.request.route_url("users.view", uuid=user.uuid)
                links.append(tags.link_to(user.username, url))
            return ", ".join(links)

        return ", ".join([user.username for user in log.owners])

    def grid_row_class(self, log, data, i):
        if log.status == "pending":
            return "has-background-warning"
        if log.status == "abandoned":
            return "has-background-danger"
        return None

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        enum = self.app.enum
        session = self.Session()
        log = f.model_instance

        # timestamp
        # TODO: the widget should be automatic (assn proxy field)
        f.set_widget("timestamp", WuttaDateTimeWidget(self.request))
        if self.creating:
            f.set_default("timestamp", self.app.make_utc())

        # assets
        f.set_node("assets", AssetRefs(self.request))
        f.set_required("assets", False)
        if not self.creating:
            # nb. must explicity declare value for non-standard field
            f.set_default("assets", log.assets)

        # groups
        f.set_node("groups", AssetRefs(self.request, is_group=True))
        if not self.creating:
            # nb. must explicity declare value for non-standard field
            f.set_default("groups", log.groups)

        # locations
        f.set_node("locations", AssetRefs(self.request, is_location=True))
        if not self.creating:
            # nb. must explicity declare value for non-standard field
            f.set_default("locations", log.locations)

        # log_type
        if self.creating:
            f.remove("log_type")
        else:
            f.set_node(
                "log_type",
                WuttaDictEnum(
                    self.request, get_log_type_enum(self.config, session=session)
                ),
            )
            f.set_readonly("log_type")

        # quantities
        f.set_node("quantities", QuantityRefs(self.request))
        if not self.creating:
            # nb. must explicity declare value for non-standard field
            f.set_default(
                "quantities", [self.app.get_true_quantity(q) for q in log.quantities]
            )

        # notes
        f.set_widget("notes", "notes")

        # owners
        if self.creating or self.editing:
            f.remove("owners")  # TODO: need to support this
        else:
            f.set_node("owners", OwnerRefs(self.request))
            # nb. must explicity declare value for non-standard field
            f.set_default("owners", log.owners)

        # status
        f.set_node("status", WuttaDictEnum(self.request, enum.LOG_STATUS))

        # is_movement
        f.set_node("is_movement", colander.Boolean())

        # is_group_assignment
        f.set_node("is_group_assignment", colander.Boolean())

        # quick
        f.set_readonly("quick")  # TODO

    def objectify(self, form):
        log = super().objectify(form)
        data = form.validated

        if self.creating:

            # log_type
            log.log_type = self.get_farmos_log_type()

            # owner
            log.owners = [self.request.user]

        self.set_assets(log, data["assets"])
        self.set_locations(log, data["locations"])
        self.set_groups(log, data["groups"])
        self.set_quantities(log, data["quantities"])

        return log

    def set_assets(self, log, desired):
        model = self.app.model
        session = self.Session()
        current = [a.uuid for a in log.assets]

        for uuid in desired:
            if uuid not in current:
                asset = session.get(model.Asset, uuid)
                assert asset
                log.assets.append(asset)

        for uuid in current:
            if uuid not in desired:
                asset = session.get(model.Asset, uuid)
                assert asset
                log.assets.remove(asset)

    def set_locations(self, log, desired):
        model = self.app.model
        session = self.Session()
        current = [l.uuid for l in log.locations]

        for uuid in desired:
            if uuid not in current:
                location = session.get(model.Asset, uuid)
                assert location
                log.locations.append(location)

        for uuid in current:
            if uuid not in desired:
                location = session.get(model.Asset, uuid)
                assert location
                log.locations.remove(location)

    def set_groups(self, log, desired):
        model = self.app.model
        session = self.Session()
        current = [g.uuid for g in log.groups]

        for uuid in desired:
            if uuid not in current:
                group = session.get(model.Asset, uuid)
                assert group
                log.groups.append(group)

        for uuid in current:
            if uuid not in desired:
                group = session.get(model.Asset, uuid)
                assert group
                log.groups.remove(group)

    def set_quantities(self, log, desired):
        model = self.app.model
        session = self.Session()

        current = {
            qty.uuid.hex: self.app.get_true_quantity(qty) for qty in log.quantities
        }
        for new_qty in desired:
            units = session.get(model.Unit, new_qty["units"]["uuid"])
            assert units
            if new_qty["uuid"].startswith("new_"):
                qty = self.app.make_true_quantity(
                    new_qty["quantity_type"]["drupal_id"],
                    measure_id=new_qty["measure"],
                    value_numerator=int(new_qty["value"]),
                    value_denominator=1,
                    units=units,
                )
                # nb. must ensure "typed" quantity record persists!
                session.add(qty)
                # but must add "generic" quantity record to log
                log.quantities.append(qty.quantity)
            else:
                old_qty = current[new_qty["uuid"]]
                old_qty.measure_id = new_qty["measure"]
                old_qty.value_numerator = int(new_qty["value"])
                old_qty.value_denominator = 1
                old_qty.units = units
                if old_qty.quantity_type_id == "material":
                    self.set_material_types(old_qty, new_qty["material_types"])

        desired = [qty["uuid"] for qty in desired]
        for old_qty in list(log.quantities):
            # nb. "old_qty" may be newly-created, w/ no uuid yet
            # (this logic may break if session gets flushed early!)
            if old_qty.uuid and old_qty.uuid.hex not in desired:
                log.quantities.remove(old_qty)

    def set_material_types(self, quantity, desired):
        model = self.app.model
        session = self.Session()
        current = {mtype.uuid: mtype for mtype in quantity.material_types}

        for new_mtype in desired:
            mtype = session.get(model.MaterialType, new_mtype["uuid"])
            assert mtype
            if mtype.uuid not in current:
                quantity.material_types.append(mtype)

        desired = [mtype["uuid"] for mtype in desired]
        for old_mtype in current.values():
            if old_mtype.uuid.hex not in desired:
                quantity.material_types.remove(old_mtype)

    def auto_sync_to_farmos(self, client, log):
        model = self.app.model
        session = self.Session()

        # nb. ensure quantities have uuid keys
        session.flush()

        for qty in log.quantities:
            qty = self.app.get_true_quantity(qty)
            self.app.auto_sync_to_farmos(qty, client=client)

        self.app.auto_sync_to_farmos(log, client=client)

    def get_farmos_url(self, log):
        return self.app.get_farmos_url(f"/log/{log.drupal_id}")

    def get_farmos_log_type(self):
        return self.model_class.__wutta_hint__["farmos_log_type"]

    def get_xref_buttons(self, log):
        buttons = super().get_xref_buttons(log)

        if log.farmos_uuid:
            log_type = self.get_farmos_log_type()
            route = f"farmos_logs_{log_type}.view"
            buttons.append(
                self.make_button(
                    "View farmOS record",
                    primary=True,
                    url=self.request.route_url(route, uuid=log.farmos_uuid),
                    icon_left="eye",
                )
            )

        return buttons

    def get_version_joins(self):
        """
        We override this to declare the relationship between the
        view's data model (which is some type of log table) and the
        canonical ``Log`` model, so the revision history views include
        transactions which reference either version table.

        See also parent method,
        :meth:`~wuttaweb:wuttaweb.views.master.MasterView.get_version_joins()`
        """
        model = self.app.model
        return super().get_version_joins() + [
            model.Log,
            (model.LogAsset, "log_uuid", "uuid"),
        ]


class AllLogView(LogMasterView):
    """
    Master view for All Logs
    """

    model_class = Log
    route_prefix = "log"
    url_prefix = "/logs"

    farmos_refurl_path = "/logs"

    viewable = False
    creatable = False
    editable = False
    deletable = False
    model_is_versioned = False

    grid_columns = [
        "status",
        "drupal_id",
        "timestamp",
        "message",
        "log_type",
        "assets",
        "locations",
        "quantities",
        "groups",
        "is_group_assignment",
        "owners",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)
        session = self.Session()

        # log_type
        g.set_enum("log_type", get_log_type_enum(self.config, session=session))

        # view action links to final log record
        def log_url(log, i):
            return self.request.route_url(f"logs_{log.log_type}.view", uuid=log.uuid)

        g.add_action("view", icon="eye", url=log_url)


def defaults(config, **kwargs):
    base = globals()

    LogTypeView = kwargs.get("LogTypeView", base["LogTypeView"])
    LogTypeView.defaults(config)

    AllLogView = kwargs.get("AllLogView", base["AllLogView"])
    AllLogView.defaults(config)


def includeme(config):
    defaults(config)



"""add MedicalLog.vet

Revision ID: d459db991404
Revises: 9e875e5cbdc1
Create Date: 2026-02-28 22:17:57.001134

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "d459db991404"
down_revision: Union[str, None] = "9e875e5cbdc1"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # log_medical
    op.add_column("log_medical", sa.Column("vet", sa.String(length=100), nullable=True))
    op.add_column(
        "log_medical_version",
        sa.Column("vet", sa.String(length=100), autoincrement=False, nullable=True),
    )


def downgrade() -> None:

    # log_medical
    op.drop_column("log_medical_version", "vet")
    op.drop_column("log_medical", "vet")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
View for farmOS animal types
"""

from wuttafarm.web.views.farmos.master import TaxonomyMasterView


class AnimalTypeView(TaxonomyMasterView):
    """
    Master view for Animal Types in farmOS.
    """

    model_name = "farmos_animal_type"
    model_title = "farmOS Animal Type"
    model_title_plural = "farmOS Animal Types"

    route_prefix = "farmos_animal_types"
    url_prefix = "/farmOS/animal-types"

    farmos_taxonomy_type = "animal_type"
    farmos_refurl_path = "/admin/structure/taxonomy/manage/animal_type/overview"

    def get_xref_buttons(self, animal_type):
        buttons = super().get_xref_buttons(animal_type)
        model = self.app.model
        session = self.Session()

        if wf_animal_type := (
            session.query(model.AnimalType)
            .filter(model.AnimalType.farmos_uuid == animal_type["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url(
                        "animal_types.view", uuid=wf_animal_type.uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    AnimalTypeView = kwargs.get("AnimalTypeView", base["AnimalTypeView"])
    AnimalTypeView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Custom app handler for WuttaFarm
"""

from wuttjamaican import app as base


class WuttaFarmAppHandler(base.AppHandler):
    """
    Custom :term:`app handler` for WuttaFarm.
    """

    display_format_datetime = "%a, %m/%d/%Y - %H:%M"

    default_auth_handler_spec = "wuttafarm.auth:WuttaFarmAuthHandler"
    default_install_handler_spec = "wuttafarm.install:WuttaFarmInstallHandler"

    def get_asset_handler(self):
        """
        Get the configured asset handler.

        :rtype: :class:`~wuttafarm.assets.AssetHandler`
        """
        if "asset" not in self.handlers:
            spec = self.config.get(
                f"{self.appname}.asset_handler",
                default="wuttafarm.assets:AssetHandler",
            )
            factory = self.load_object(spec)
            self.handlers["asset"] = factory(self.config)
        return self.handlers["asset"]

    def get_farmos_handler(self):
        """
        Get the configured farmOS integration handler.

        :rtype: :class:`~wuttafarm.farmos.FarmOSHandler`
        """
        if "farmos" not in self.handlers:
            spec = self.config.get(
                f"{self.appname}.farmos_handler",
                default="wuttafarm.farmos.handler:FarmOSHandler",
            )
            factory = self.load_object(spec)
            self.handlers["farmos"] = factory(self.config)
        return self.handlers["farmos"]

    def get_farmos_integration_mode(self):
        """
        Returns the integration mode for farmOS, i.e. to control the
        app's behavior regarding that.
        """
        enum = self.enum
        return self.config.get(
            f"{self.appname}.farmos_integration_mode",
            default=enum.FARMOS_INTEGRATION_MODE_WRAPPER,
        )

    def is_farmos_mirror(self):
        """
        Returns ``True`` if the app is configured in "mirror"
        integration mode with regard to farmOS.
        """
        enum = self.enum
        mode = self.get_farmos_integration_mode()
        return mode == enum.FARMOS_INTEGRATION_MODE_MIRROR

    def is_farmos_wrapper(self):
        """
        Returns ``True`` if the app is configured in "wrapper"
        integration mode with regard to farmOS.
        """
        enum = self.enum
        mode = self.get_farmos_integration_mode()
        return mode == enum.FARMOS_INTEGRATION_MODE_WRAPPER

    def is_standalone(self):
        """
        Returns ``True`` if the app is configured in "standalone" mode
        with regard to farmOS.
        """
        enum = self.enum
        mode = self.get_farmos_integration_mode()
        return mode == enum.FARMOS_INTEGRATION_MODE_NONE

    def get_farmos_url(self, *args, **kwargs):
        """
        Get a farmOS URL.  This is a convenience wrapper around
        :meth:`~wuttafarm.farmos.handler.FarmOSHandler.get_farmos_url()`.
        """
        handler = self.get_farmos_handler()
        return handler.get_farmos_url(*args, **kwargs)

    def get_farmos_client(self, *args, **kwargs):
        """
        Get a farmOS client.  This is a convenience wrapper around
        :meth:`~wuttafarm.farmos.handler.FarmOSHandler.get_farmos_client()`.
        """
        handler = self.get_farmos_handler()
        return handler.get_farmos_client(*args, **kwargs)

    def is_farmos_3x(self, *args, **kwargs):
        """
        Check if the farmOS version is 3.x.  This is a convenience
        wrapper around
        :meth:`~wuttafarm.farmos.handler.FarmOSHandler.is_farmos_3x()`.
        """
        handler = self.get_farmos_handler()
        return handler.is_farmos_3x(*args, **kwargs)

    def is_farmos_4x(self, *args, **kwargs):
        """
        Check if the farmOS version is 4.x.  This is a convenience
        wrapper around
        :meth:`~wuttafarm.farmos.handler.FarmOSHandler.is_farmos_4x()`.
        """
        handler = self.get_farmos_handler()
        return handler.is_farmos_4x(*args, **kwargs)

    def get_normalizer(self, farmos_client=None):
        """
        Get the configured farmOS integration handler.

        :rtype: :class:`~wuttafarm.farmos.FarmOSHandler`
        """
        spec = self.config.get(
            f"{self.appname}.normalizer_spec",
            default="wuttafarm.normal:Normalizer",
        )
        factory = self.load_object(spec)
        return factory(self.config, farmos_client)

    def get_quantity_types(self, session=None):
        """
        Returns a list of all known quantity types.
        """
        model = self.model
        with self.short_session(session=session) as sess:
            return (
                sess.query(model.QuantityType).order_by(model.QuantityType.name).all()
            )

    def get_measures(self, session=None):
        """
        Returns a list of all known measures.
        """
        model = self.model
        with self.short_session(session=session) as sess:
            return sess.query(model.Measure).order_by(model.Measure.ordinal).all()

    def get_units(self, session=None):
        """
        Returns a list of all known units.
        """
        model = self.model
        with self.short_session(session=session) as sess:
            return sess.query(model.Unit).order_by(model.Unit.name).all()

    def get_material_types(self, session=None):
        """
        Returns a list of all known material types.
        """
        model = self.model
        with self.short_session(session=session) as sess:
            return (
                sess.query(model.MaterialType).order_by(model.MaterialType.name).all()
            )

    def get_quantity_models(self):
        model = self.model
        return {
            "standard": model.StandardQuantity,
            "material": model.MaterialQuantity,
        }

    def get_true_quantity(self, quantity, require=True):
        model = self.model
        if not isinstance(quantity, model.Quantity):
            if require and not quantity:
                raise ValueError(f"quantity is not valid: {quantity}")
            return quantity

        session = self.get_session(quantity)
        models = self.get_quantity_models()
        if require and quantity.quantity_type_id not in models:
            raise ValueError(
                f"quantity has invalid quantity_type_id: {quantity.quantity_type_id}"
            )

        true_quantity = session.get(models[quantity.quantity_type_id], quantity.uuid)
        if require and not true_quantity:
            raise ValueError(f"quantity has no true/typed quantity record: {quantity}")

        return true_quantity

    def make_true_quantity(self, quantity_type_id, **kwargs):
        models = self.get_quantity_models()
        kwargs["quantity_type_id"] = quantity_type_id
        return models[quantity_type_id](**kwargs)

    def auto_sync_to_farmos(self, obj, model_name=None, client=None, require=True):
        """
        Export the given object to farmOS, using configured handler.

        This should ensure the given object is also *updated* with the
        farmOS UUID and Drupal ID, when new record is created in
        farmOS.

        :param obj: Any data object in WuttaFarm, e.g. AnimalAsset
           instance.

        :param client: Existing farmOS API client to use.  If not
           specified, a new one will be instantiated.

        :param require: If true, this will *require* the export
           handler to support objects of the given type.  If false,
           then nothing will happen / export is silently skipped when
           there is no such exporter.
        """
        handler = self.app.get_import_handler("export.to_farmos.from_wuttafarm")

        if not model_name:
            model_name = type(obj).__name__
        if model_name not in handler.importers:
            if require:
                raise ValueError(f"no exporter found for {model_name}")
            return

        # nb. begin txn to establish the API client
        handler.begin_target_transaction(client)
        importer = handler.get_importer(model_name, caches_target=False)
        normal = importer.normalize_source_object(obj)
        importer.process_data(source_data=[normal])

    def auto_sync_from_farmos(self, obj, model_name, client=None, require=True):
        """
        Import the given object from farmOS, using configured handler.

        :param obj: Any data record from farmOS.

        :param model_name': Model name for the importer to use,
           e.g. ``"AnimalAsset"``.

        :param client: Existing farmOS API client to use.  If not
           specified, a new one will be instantiated.

        :param require: If true, this will *require* the import
           handler to support objects of the given type.  If false,
           then nothing will happen / import is silently skipped when
           there is no such importer.
        """
        handler = self.app.get_import_handler("import.to_wuttafarm.from_farmos")

        if model_name not in handler.importers:
            if require:
                raise ValueError(f"no importer found for {model_name}")
            return

        # nb. begin txn to establish the API client
        handler.begin_source_transaction(client)
        with self.short_session(commit=True) as session:
            handler.target_session = session
            importer = handler.get_importer(model_name, caches_target=False)
            normal = importer.normalize_source_object(obj)
            importer.process_data(source_data=[normal])


class WuttaFarmAppProvider(base.AppProvider):
    """
    The :term:`app provider` for WuttaFarm.
    """

    email_modules = ["wuttafarm.emails"]



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Custom form widgets for WuttaFarm
"""

import json

import colander
from deform.widget import Widget, SelectWidget, sequence_types, _normalize_choices
from webhelpers2.html import HTML, tags

from wuttaweb.forms.widgets import WuttaCheckboxChoiceWidget, ObjectRefWidget
from wuttaweb.db import Session

from wuttafarm.web.util import render_quantity_objects
from wuttafarm.db.model import EquipmentType


class ImageWidget(Widget):
    """
    Widget to display an image URL for a record.
    """

    def __init__(self, alt_text, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alt_text = alt_text

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            return tags.image(cstruct, self.alt_text, **kw)

        return super().serialize(field, cstruct, **kw)


class LogQuickWidget(Widget):
    """
    Widget to display an image URL for a record.
    """

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            items = []
            for quick in json.loads(cstruct):
                items.append(HTML.tag("li", c=quick))
            return HTML.tag("ul", c=items)

        return super().serialize(field, cstruct, **kw)


class FarmOSRefWidget(SelectWidget):
    """
    Generic widget to display "any reference field" - as a link to
    view the farmOS record it references.  Only used by the farmOS
    direct API views.
    """

    def __init__(self, request, route_prefix, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.route_prefix = route_prefix

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            try:
                obj = json.loads(cstruct)
            except json.JSONDecodeError:
                name = dict(self.values)[cstruct]
                obj = {"uuid": cstruct, "name": name}

            return tags.link_to(
                obj["name"],
                self.request.route_url(f"{self.route_prefix}.view", uuid=obj["uuid"]),
            )

        return super().serialize(field, cstruct, **kw)


class FarmOSRefsWidget(Widget):

    def __init__(self, request, route_prefix, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.route_prefix = route_prefix

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            links = []
            for obj in cstruct:
                url = self.request.route_url(
                    f"{self.route_prefix}.view", uuid=obj["uuid"]
                )
                links.append(HTML.tag("li", c=tags.link_to(obj["name"], url)))

            return HTML.tag("ul", c=links)

        return super().serialize(field, cstruct, **kw)


class FarmOSAssetRefsWidget(Widget):
    """
    Widget to display a "Assets" field for an asset.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            assets = []
            for asset in json.loads(cstruct):
                url = self.request.route_url(
                    f"farmos_{asset['asset_type']}_assets.view", uuid=asset["uuid"]
                )
                assets.append(HTML.tag("li", c=tags.link_to(asset["name"], url)))

            return HTML.tag("ul", c=assets)

        return super().serialize(field, cstruct, **kw)


class FarmOSLocationRefsWidget(Widget):
    """
    Widget to display a "Locations" field for an asset.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            locations = []
            for location in json.loads(cstruct):
                asset_type = location["type"].split("--")[1]
                url = self.request.route_url(
                    f"farmos_{asset_type}_assets.view", uuid=location["uuid"]
                )
                locations.append(HTML.tag("li", c=tags.link_to(location["name"], url)))

            return HTML.tag("ul", c=locations)

        return super().serialize(field, cstruct, **kw)


class FarmOSQuantityRefsWidget(Widget):
    """
    Widget to display a "Quantities" field for a log.
    """

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            quantities = json.loads(cstruct)
            return render_quantity_objects(quantities)

        return super().serialize(field, cstruct, **kw)


class FarmOSUnitRefWidget(Widget):
    """
    Widget to display a "Units" field for a quantity.
    """

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            unit = json.loads(cstruct)
            return unit["name"]

        return super().serialize(field, cstruct, **kw)


class FarmOSTaxonomyTermsWidget(Widget):
    """
    Widget to display a field which can reference multiple taxonomy
    terms.
    """

    def __init__(self, request, route_prefix, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.route_prefix = route_prefix

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            links = []
            for term in cstruct:
                link = tags.link_to(
                    term["name"],
                    self.request.route_url(
                        f"{self.route_prefix}.view", uuid=term["uuid"]
                    ),
                )
                links.append(HTML.tag("li", c=link))
            return HTML.tag("ul", c=links)

        return super().serialize(field, cstruct, **kw)


class FarmOSPlantTypesWidget(Widget):
    """
    Widget to display a farmOS "plant types" field.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            links = []
            for plant_type in json.loads(cstruct):
                link = tags.link_to(
                    plant_type["name"],
                    self.request.route_url(
                        "farmos_plant_types.view", uuid=plant_type["uuid"]
                    ),
                )
                links.append(HTML.tag("li", c=link))
            return HTML.tag("ul", c=links)

        return super().serialize(field, cstruct, **kw)


class TaxonomyTermRefsWidget(Widget):
    """
    Generic (incomplete) widget for fields which can reference
    multiple taxonomy terms.

    This widget can handle typical read-only scenarios but the
    editable mode is not implemented.
    """

    route_prefix = None

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    @classmethod
    def get_route_prefix(cls):
        return cls.route_prefix

    @classmethod
    def get_permission_prefix(cls):
        return cls.route_prefix

    def serialize(self, field, cstruct, **kw):
        """ """
        if not cstruct:
            cstruct = []

        if readonly := kw.get("readonly", self.readonly):
            items = []
            route_prefix = self.get_route_prefix()
            for term in cstruct:
                url = self.request.route_url(f"{route_prefix}.view", uuid=term["uuid"])
                link = tags.link_to(term["name"], url)
                items.append(HTML.tag("li", c=link))
            return HTML.tag("ul", c=items)

        tmpl_values = self.get_template_values(field, cstruct, kw)
        return field.renderer(self.template, **tmpl_values)

    def get_template_values(self, field, cstruct, kw):
        values = super().get_template_values(field, cstruct, kw)
        model = self.app.model
        session = Session()

        terms = []
        query = session.query(self.model_class).order_by(self.model_class.name)
        for term in query:
            terms.append(
                {
                    "uuid": str(term.uuid),
                    "name": term.name,
                }
            )
        values["terms"] = terms

        permission_prefix = self.get_permission_prefix()
        if self.request.has_perm(f"{permission_prefix}.create"):
            values["can_create"] = True

        return values

    def deserialize(self, field, pstruct):
        """ """
        if not pstruct:
            return colander.null

        return json.loads(pstruct)


class EquipmentTypeRefsWidget(TaxonomyTermRefsWidget):
    """
    Widget for Equipment Types field.
    """

    model_class = EquipmentType
    route_prefix = "equipment_types"
    template = "equipmenttyperefs"


class PlantTypeRefsWidget(Widget):
    """
    Widget for Plant Types field (on a Plant Asset).
    """

    template = "planttyperefs"
    values = ()

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, field, cstruct, **kw):
        """ """
        model = self.app.model
        session = Session()

        if cstruct in (colander.null, None):
            cstruct = ()

        if readonly := kw.get("readonly", self.readonly):
            items = []

            plant_types = (
                session.query(model.PlantType)
                .filter(model.PlantType.uuid.in_(cstruct))
                .order_by(model.PlantType.name)
                .all()
            )

            for plant_type in plant_types:
                items.append(
                    HTML.tag(
                        "li",
                        c=tags.link_to(
                            str(plant_type),
                            self.request.route_url(
                                "plant_types.view", uuid=plant_type.uuid
                            ),
                        ),
                    )
                )

            return HTML.tag("ul", c=items)

        values = kw.get("values", self.values)
        if not isinstance(values, sequence_types):
            raise TypeError("Values must be a sequence type (list, tuple, or range).")

        kw["values"] = _normalize_choices(values)
        tmpl_values = self.get_template_values(field, cstruct, kw)
        return field.renderer(self.template, **tmpl_values)

    def get_template_values(self, field, cstruct, kw):
        """ """
        values = super().get_template_values(field, cstruct, kw)

        values["js_values"] = json.dumps(values["values"])

        if self.request.has_perm("plant_types.create"):
            values["can_create"] = True

        return values

    def deserialize(self, field, pstruct):
        """ """
        if not pstruct:
            return colander.null

        return set(pstruct.split(","))


class MaterialTypeRefsWidget(Widget):
    """
    Widget for Material Types field (on a Material Asset).
    """

    template = "materialtyperefs"
    values = ()

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, field, cstruct, **kw):
        """ """
        model = self.app.model
        session = Session()

        if not cstruct:
            cstruct = []

        if readonly := kw.get("readonly", self.readonly):
            items = []
            for mtype in cstruct:
                items.append(
                    HTML.tag(
                        "li",
                        c=tags.link_to(
                            mtype["name"],
                            self.request.route_url(
                                "material_types.view", uuid=mtype["uuid"]
                            ),
                        ),
                    )
                )
            return HTML.tag("ul", c=items)

        tmpl_values = self.get_template_values(field, cstruct, kw)
        return field.renderer(self.template, **tmpl_values)

    def get_template_values(self, field, cstruct, kw):
        """ """
        values = super().get_template_values(field, cstruct, kw)
        session = Session()

        material_types = []
        for mtype in self.app.get_material_types(session):
            material_types.append(
                {
                    "uuid": mtype.uuid.hex,
                    "name": mtype.name,
                }
            )
        values["material_types"] = json.dumps(material_types)

        return values

    def deserialize(self, field, pstruct):
        """ """
        if not pstruct:
            return []

        return json.loads(pstruct)


class SeasonRefsWidget(Widget):
    """
    Widget for Seasons field (on a Plant Asset).
    """

    template = "seasonrefs"
    values = ()

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, field, cstruct, **kw):
        """ """
        model = self.app.model
        session = Session()

        if cstruct in (colander.null, None):
            cstruct = ()

        if readonly := kw.get("readonly", self.readonly):
            items = []

            seasons = (
                session.query(model.Season)
                .filter(model.Season.uuid.in_(cstruct))
                .order_by(model.Season.name)
                .all()
            )

            for season in seasons:
                items.append(
                    HTML.tag(
                        "li",
                        c=tags.link_to(
                            str(season),
                            self.request.route_url("seasons.view", uuid=season.uuid),
                        ),
                    )
                )

            return HTML.tag("ul", c=items)

        values = kw.get("values", self.values)
        if not isinstance(values, sequence_types):
            raise TypeError("Values must be a sequence type (list, tuple, or range).")

        kw["values"] = _normalize_choices(values)
        tmpl_values = self.get_template_values(field, cstruct, kw)
        return field.renderer(self.template, **tmpl_values)

    def get_template_values(self, field, cstruct, kw):
        """ """
        values = super().get_template_values(field, cstruct, kw)

        values["js_values"] = json.dumps(values["values"])

        if self.request.has_perm("seasons.create"):
            values["can_create"] = True

        return values

    def deserialize(self, field, pstruct):
        """ """
        if not pstruct:
            return set()

        return set(pstruct.split(","))


class StructureWidget(Widget):
    """
    Widget to display a "structure" field.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            structure = json.loads(cstruct)
            return tags.link_to(
                structure["name"],
                self.request.route_url(
                    "farmos_structure_assets.view", uuid=structure["uuid"]
                ),
            )

        return super().serialize(field, cstruct, **kw)


class UsersWidget(Widget):
    """
    Widget to display the list of owners for an asset etc.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            items = []
            for user in json.loads(cstruct):
                link = tags.link_to(
                    user["display_name"],
                    self.request.route_url("farmos_users.view", uuid=user["uuid"]),
                )
                items.append(HTML.tag("li", c=link))

            return HTML.tag("ul", c=items)

        return super().serialize(field, cstruct, **kw)


##############################
# native data widgets
##############################


class AssetRefsWidget(Widget):
    """
    Widget for Assets field (of various kinds).
    """

    template = "assetrefs"
    values = ()

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, field, cstruct, **kw):
        """ """
        model = self.app.model
        session = Session()

        readonly = kw.get("readonly", self.readonly)
        if readonly:
            assets = []
            for uuid in cstruct or []:
                asset = session.get(model.Asset, uuid)
                assets.append(
                    HTML.tag(
                        "li",
                        c=tags.link_to(
                            str(asset),
                            self.request.route_url(
                                f"{asset.asset_type}_assets.view", uuid=asset.uuid
                            ),
                        ),
                    )
                )
            return HTML.tag("ul", c=assets)

        values = kw.get("values", self.values)
        if not isinstance(values, sequence_types):
            raise TypeError("Values must be a sequence type (list, tuple, or range).")

        kw["values"] = _normalize_choices(values)
        tmpl_values = self.get_template_values(field, cstruct, kw)
        return field.renderer(self.template, **tmpl_values)

    def get_template_values(self, field, cstruct, kw):
        """ """
        values = super().get_template_values(field, cstruct, kw)

        values["js_values"] = json.dumps(values["values"])

        return values

    def deserialize(self, field, pstruct):
        """ """
        if not pstruct:
            return set()

        return set(pstruct.split(","))


class QuantityRefsWidget(Widget):
    """
    Widget for Quantities field (on a Log record)
    """

    template = "quantityrefs"

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, field, cstruct, **kw):
        """ """
        model = self.app.model
        session = Session()

        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if not cstruct:
                return ""

            quantities = []
            for qty in cstruct:
                url = self.request.route_url(
                    f"quantities_{qty['quantity_type']['drupal_id']}.view",
                    uuid=qty["uuid"],
                )
                quantities.append(HTML.tag("li", c=tags.link_to(qty["as_text"], url)))

            return HTML.tag("ul", c=quantities)

        tmpl_values = self.get_template_values(field, cstruct, kw)
        return field.renderer(self.template, **tmpl_values)

    def get_template_values(self, field, cstruct, kw):
        model = self.app.model
        session = Session()
        values = super().get_template_values(field, cstruct, kw)

        qtypes = []
        for qtype in self.app.get_quantity_types(session):
            qtypes.append(
                {
                    "uuid": qtype.uuid.hex,
                    "drupal_id": qtype.drupal_id,
                    "name": qtype.name,
                }
            )
        values["quantity_types"] = qtypes

        material_types = []
        for mtype in self.app.get_material_types(session):
            material_types.append(
                {
                    "uuid": mtype.uuid.hex,
                    "name": mtype.name,
                }
            )
        values["material_types"] = material_types

        measures = []
        for measure in self.app.get_measures(session):
            measures.append(
                {
                    "uuid": measure.uuid.hex,
                    "drupal_id": measure.drupal_id,
                    "name": measure.name,
                }
            )
        values["measures"] = measures

        units = []
        for unit in self.app.get_units(session):
            units.append(
                {
                    "uuid": unit.uuid.hex,
                    "drupal_id": unit.drupal_id,
                    "name": unit.name,
                }
            )
        values["units"] = units

        return values

    def deserialize(self, field, pstruct):
        """ """
        if not pstruct:
            return set()

        return json.loads(pstruct)


class OwnerRefsWidget(WuttaCheckboxChoiceWidget):
    """
    Widget for Owners field (on an Asset or Log record)
    """

    def serialize(self, field, cstruct, **kw):
        """ """
        model = self.app.model
        session = Session()

        readonly = kw.get("readonly", self.readonly)
        if readonly:
            owners = [session.get(model.User, uuid) for uuid in cstruct or []]
            owners = [user for user in owners if user]
            owners.sort(key=lambda user: user.username)
            links = []
            for user in owners:
                links.append(
                    HTML.tag(
                        "li",
                        c=tags.link_to(
                            user.username,
                            self.request.route_url("users.view", uuid=user.uuid),
                        ),
                    )
                )
            return HTML.tag("ul", c=links)

        return super().serialize(field, cstruct, **kw)


class AnimalTypeRefWidget(ObjectRefWidget):
    """
    Custom widget which uses the ``<animal-type-picker>`` component.
    """

    template = "animaltyperef"

    def get_template_values(self, field, cstruct, kw):
        """ """
        values = super().get_template_values(field, cstruct, kw)

        values["js_values"] = json.dumps(values["values"])

        if self.request.has_perm("animal_types.create"):
            values["can_create"] = True

        return values
